import os
import sys
import time
import copy
from typing import List, Optional, TextIO, Tuple

import yaml

from vessl.openapi_client import ModelServiceLinkSecretAPIInput
from vessl.openapi_client.models import (
    ModelServiceCreateAPIInput,
    ModelserviceModelServiceListResponse,
    ModelserviceModelServiceReadResponse,
    ResponseModelServiceInfo,
    ResponseModelServiceRevision,
    ResponseSecret,
    V1HubModelTaskSpec,
)
from vessl.openapi_client.models.v1_autoscaling import V1Autoscaling
from vessl.openapi_client.models.v1_auxiliary_form import V1AuxiliaryForm
from vessl.openapi_client.models.v1_command import V1Command
from vessl.openapi_client.models.v1_env_form import V1EnvForm
from vessl.openapi_client.models.v1_run_argument import V1RunArgument
from vessl.openapi_client.models.v1_import_form import V1ImportForm
from vessl.openapi_client.models.v1_model_ref_form import V1ModelRefForm
from vessl.openapi_client.models.v1_port_path import V1PortPath
from vessl.openapi_client.models.v1_resources_form import V1ResourcesForm
from vessl.openapi_client.models.v1_serve_revision_service_form import (
    V1ServeRevisionServiceForm,
)
from vessl import vessl_api
from vessl.cli._util import (
    print_table_tabulate,
)
from vessl.util.prompt import prompt_confirm, prompt_text, prompt_choices
from vessl.util.echo import print_debug, print_info, print_success, print_error, print_error_result
from vessl.util.endpoint import Endpoint
from vessl.cli.serve.util import (
    print_gateway,
    print_services,
    print_revision,
    create_secret,
    fetch_variable_from_injects,
    get_runner_type,
    sanitize,
    sanitize_model_form,
    generate_image_name
)
from vessl.enums import ModelServiceType, ClusterCustomEndpointType
from vessl.models.vessl_model import VesslModel
from vessl.kernel_cluster import list_cluster_presets, list_clusters
from vessl.organization import _get_organization_name
from vessl.secret import list_generic_secrets
from vessl.serving import (
    _wait_for_gateway_enabled,
    _wait_for_revision_to_launch,
    abort_in_progress_rollout,
    abort_in_progress_rollout_by_name,
    create_active_revision_replacement_rollout,
    create_revision_from_yaml_v2,
    get_recent_rollout,
    launch_revision,
    list_services,
    _read_service,
    update_model_service,
    update_revision_autoscaler_config,
)

from vessl.util.exception import VesslApiException

from .command_options import build_autoscaling_configs


def list():
    services: List[ResponseModelServiceInfo] = list_services(
        organization=vessl_api.organization.name
    ).results
    print_services(services=services)

def abort_update(service: ResponseModelServiceInfo):
    try:
        abort_in_progress_rollout(service=service)
    except VesslApiException as e:
        print(f"Failed to abort the service update: {e.message}")
        sys.exit(1)

def create(
    file: TextIO,
    no_prompt: bool,
    launch: bool,
    set_current_active: bool,
    force: bool,
    service_name: str,
    serverless: bool,
    from_hub: str,
    node_port: int,
    subdomain: str,
):
    organization_name = vessl_api.set_organization()
    vessl_api.configure_default_organization(organization_name)

    if from_hub:
        hub: V1HubModelTaskSpec = vessl_api.hub_model_task_read_by_key_api(key=from_hub)
        yaml_body = hub.yaml
    elif file:
        yaml_body = file.read()
    else:
        print_error("Error: Either --from-hub or --file must be specified.")
        sys.exit(1)

    try:
        yaml_loaded = yaml.safe_load(yaml_body)
    except yaml.YAMLError as e:
        print_error(f"Error: invalid YAML\n{e}")
        sys.exit(1)

    if not isinstance(yaml_loaded, dict):
        print_error(f"Error: invalid YAML: expected mapping (dict), got {type(yaml_loaded)}")
        sys.exit(1)

    try:
        cluster_name = yaml_loaded["resources"]["cluster"]
    except (KeyError, TypeError):
        print_error("Error: invalid YAML: no cluster name")
        sys.exit(1)
    
    arguments = {}
    if from_hub:
        injects = vessl_api.hub_model_injects_read_by_key_api(key=from_hub)

        if not no_prompt and injects.injected_envs:
            for key in injects.injected_envs:
                arguments[key] = fetch_variable_from_injects(key)

    service_name_from_legacy_yaml = yaml_loaded.get("name", None)

    service_name = service_name or service_name_from_legacy_yaml or None
    service_name, service_type = _check_service_name_or_prompt(
        service_name=service_name,
        cluster_name=cluster_name,
        serverless=serverless,
        no_prompt=no_prompt,
        node_port=node_port,
        subdomain=subdomain,
    )

    ## Check if the service is already rolling out
    ongoing_rollout_exists = False
    try:
        recent_rollout = get_recent_rollout(service_name)
        if recent_rollout and recent_rollout.status == "rolling_out":
            ongoing_rollout_exists = True
            print_error(f"Error: the service {service_name} is currently rolling out.")
            if not force:
                print_error("Use --force option to abort the existing rollout.")
                sys.exit(1)
    except VesslApiException as e:
        print_debug("No existing rollout found.")
        pass

    ## Abort the existing rollout if --force
    if ongoing_rollout_exists and force:
        if abort_in_progress_rollout_by_name(service_name):
            print_info("Waiting for the existing rollout to be aborted...")
            time.sleep(30)

    ## create revision
    try:
        revision = create_revision_from_yaml_v2(
            organization=vessl_api.organization.name,
            service_name=service_name,
            yaml_body=yaml_body,
            serverless=(service_type == ModelServiceType.SERVERLESS),
            arguments=V1RunArgument(env_vars=arguments)
        )
    except VesslApiException as e:
        if e.status == 400:
            print_error("Error: failed to create revision (invalid parameters).")
            print_error_result(e.message)
        else:
            print_error("Error: failed to create revision. (internal error).")
        sys.exit(1)

    service_url = Endpoint.service.format(vessl_api.organization.name, service_name)
    if service_type == ModelServiceType.SERVERLESS:
        print_success(f"Successfully created revision #{revision.number}!")
        print()
        print_success(f"Check out the service at: {service_url}")
        print()
        return

    ## launch or set-active
    if set_current_active:
        launch = True
    if not launch and not no_prompt:
        launch = prompt_confirm("Do you want to launch this revision immediately?")
    if launch:
        if not set_current_active and not no_prompt:
            set_current_active = prompt_confirm("Do you want this revision to receive full traffic immediately?")

    try:
        if set_current_active:
            create_active_revision_replacement_rollout(
                organization=vessl_api.organization.name,
                model_service_name=revision.model_service_name,
                desired_active_revisions_to_weight_map={revision.number: 100},
            )
            print_info("Successfully triggered revision activation.")
        elif launch:
            launch_revision(
                organization=vessl_api.organization.name,
                service_name=revision.model_service_name,
                revision_number=revision.number,
            )
            print_info("Successfully triggered revision launch.")
        if not no_prompt and (launch or set_current_active):
            _wait_for_revision_to_launch(service_name=service_name, revision_number=revision.number, print_output=True)
        revision = vessl_api.model_service_revision_read_api(
        organization_name=vessl_api.organization.name,
        model_service_name=service_name,
        revision_number=revision.number,
        )
        print_success(f"Successfully created revision for service: {service_name}!")
        print_revision(revision, verbose=False)
        print()
        print(f"Check out the service at: {service_url}")

        
    except VesslApiException as e:
        print_error("Error: failed to launch revision.")
        print_error_result(e.message)
        sys.exit(1)

    ## Wait for the gateway to be enabled
    if set_current_active:
        gateway = _read_service(service_name=service_name).gateway_config
        if not no_prompt:
            _wait_for_gateway_enabled(gateway=gateway, service_name=revision.model_service_name, print_output=True)

        print_info("Endpoint is enabled.")
        gateway = _read_service(service_name=service_name).gateway_config
        print_gateway(gateway)
        print_info(f"You can test your service via {gateway.endpoint}")

def scale(
    service: ResponseModelServiceInfo,
    number: ResponseModelServiceRevision,
    min: Optional[int],
    max: Optional[int],
    target: Optional[int],
    metric: Optional[str],
):
    if service.type == ModelServiceType.SERVERLESS.value:
        print("User cannot update autoscaling for serverless service.")
        print("Autoscaling will be automatically set based on number of requests.")
        sys.exit(1)

    revision = number
    conf: V1Autoscaling = revision.revision_spec.autoscaling
    original_conf: V1Autoscaling = copy.copy(conf)
    updated_conf: V1Autoscaling = build_autoscaling_configs(conf=conf, min=min, max=max, target=target, metric=metric)
    
    try:
        update_revision_autoscaler_config(
            organization=vessl_api.organization.name,
            service_name=service.name,
            revision_number=revision.number,
            autoscaling=updated_conf,
        )
    except VesslApiException as e:
        print(f"Failed to update autoscaler config of revision #{revision.number} of service {service.name}: {e.message}")
        sys.exit(1)

    print("Successfully updated autoscaler config")
    print_table_tabulate([
        {
            "Original Autoscaling Configurations": original_conf,
            "Updated Autoscaling Configurations": updated_conf
        }
    ])

def split(service: ResponseModelServiceInfo, number: List[int], weight: List[int], interactive: bool):
    if not interactive:
        update_model_service(service, number, weight)
        return
    
    gateway_current = _read_service(service_name=service.name).gateway_config
    endpoint = gateway_current.endpoint
    print_info(f"Current Endpoint Host: {'(not set)' if endpoint=='' else endpoint}")
    print_info("Traffic Rules")
    if gateway_current.rules:
        current_rules = [
            {
                "Revision Number": i.revision_number,
                "Traffic Weight": f'{i.traffic_weight}%',
            }
            for i in gateway_current.rules
        ]
        print_table_tabulate(current_rules)

    print_table_tabulate(
        [
            {
                "Revision Number": n,
                "Traffic Weight": f'{w}%',
            } for n, w in zip(number, weight)
        ],
    )
    update_model_service(service, number, weight)

def set_access_token(
    service: ResponseModelServiceInfo,
    secret: ResponseSecret,
):
    vessl_api.model_service_link_secret_api(
        organization_name=vessl_api.organization.name,
        model_service_name=service.name,
        model_service_link_secret_api_input=ModelServiceLinkSecretAPIInput(
            secret_id=secret.credentials_id,
        )
    )

def create_yaml(
    service_name: str,
    message: str,
    use_api_key: bool = False,
):
    organization_name = _get_organization_name()

    lockfile_path = ".vessl.model.lock"
    lockfile = VesslModel.from_lockfile(lockfile_path)
    if lockfile is None:
        print_error("The lockfile is not found.")
        return

    try:
        msr = vessl_api.model_service_read_api(
            organization_name=organization_name,
            model_service_name=service_name,
        )
    except Exception as e:
        if e.status == 404:
            print_error(f"Model service {service_name} not found. You should first create a service.")
            return
        print_error(f"Failed to read model service: {e}")
        return

    print_info(f"Service name of {service_name} found.")

    # clusters = list_clusters()
    # cluster_name = prompt_choices("Cluster", [x.name for x in clusters])
    # cluster = list(filter(lambda x: x.name == cluster_name, clusters))[0]
    cluster = msr.kernel_cluster
    print_info(f"Using {cluster.name} cluster configured by the service.")

    presets = list_cluster_presets(cluster.id)
    preset_name = prompt_choices("Preset", [x.name for x in presets])
    preset = list(filter(lambda x: x.name == preset_name, presets))[0]

    secret_name = None
    if use_api_key:
        print_info("Select API key for authentication.")
        secrets = list_generic_secrets()
        if len(secrets) == 0:
            print_info("No secrets registered.")
            if prompt_confirm("Do you want to create a new secret?", default=True):
                secret_name = create_secret()
            else:
                print_error("No secret selected.")
                return
        else:
            secret_names = [x.credentials_name for x in secrets]
            secret_names.append("Create a new secret")
            secret_name = prompt_choices("Secret", secret_names)
            if secret_name == "Create a new secret":
                secret_name = create_secret()

    service = V1ServeRevisionServiceForm(
        expose=3000,
        healthcheck=V1PortPath(path="/",port=3000),
        monitoring=[V1PortPath(path="/metrics",port=3000)],
        autoscaling=V1Autoscaling(min=1, max=2, metric="cpu", target=50),
        auxiliary=V1AuxiliaryForm(
            runner_type=get_runner_type(lockfile.type),
        )
    ).to_dict()

    form = {
        "name":service_name,
        "message":message,
        "env":(
            None
            if use_api_key is None or secret_name is None
            else {
                "SERVICE_AUTH_KEY": V1EnvForm(
                    secret=secret_name,
                    source="secret",
                ).to_dict()
            }
        ),
        "image": generate_image_name(lockfile),
        "resources": V1ResourcesForm(
            cluster=cluster.name,
            preset=preset.name,
        ).to_dict(),
        "import": {
            "/model": sanitize_model_form(V1ImportForm(
                model=V1ModelRefForm(
                    organization_name=organization_name,
                    model_repository_name=lockfile.repository_name,
                    model_number=lockfile.model_number,
                )
            ))
        },
        "run": [
            V1Command(
                command=lockfile.entrypoint,
                workdir="/model/"+os.getcwd().split("/")[-1],
            ).to_dict()
        ],
        "ports": [V1PortPath(port=3000).to_dict()],
        "service":service,
    }

    with open("service.yaml", "w") as f:
        cleaned = sanitize(form)
        b = yaml.dump(cleaned, default_flow_style=False, sort_keys=False)
        b = b.replace("_import", "import")
        f.write(b)
    
    print_success("service.yaml created.")


def _check_service_name_or_prompt(
    service_name: Optional[str],
    cluster_name: str,
    serverless: bool,
    no_prompt: bool,
    node_port: int,
    subdomain: str,
) -> Tuple[str, ModelServiceType]:
    """
    Verifies (or prompts the user to choose) a service name.

    Arguments:
        service_name (str | None):
            Name of service, provided either in CLI argument or embedded in YAML.

        cluster_name (str):
            Name of the cluster, hinted in YAML. This is used as when we have to
            create a service.

        serverless (bool):
            Whether user has explicitly requested the service to be in serverless mode.

        no_prmopt (bool):
            True if the user has explicitly denied prompts or interactions.

        node_port (int):
            Cluster NodePort number to expose the service.

        subdomain (str):
            Subdomain to expose the service.

    Returns:
        str: Name of the selected Service.
        ModelServiceType: Type of the selected Service.
    """

    if service_name:
        try:
            model_service_resp: ModelserviceModelServiceReadResponse = (
                vessl_api.model_service_read_api(
                    model_service_name=service_name,
                    organization_name=vessl_api.organization.name,
                )
            )
        except VesslApiException as e:
            if e.status == 404:
                print_info(f"Service '{service_name}' not found.")
                print_info("Automatically creating one...")
                print()
                return _create_service(
                    service_name=service_name, cluster_name=cluster_name, serverless=serverless, node_port=node_port, subdomain=subdomain
                )
            else:  # unexpected error
                raise e

        service_type = ModelServiceType(model_service_resp.type)
        if serverless and service_type != ModelServiceType.SERVERLESS:
            print_error(
                f"Error: service '{service_name}' has type {str(service_type.value).capitalize()}, "
                "but --serverless flag was given in the command line.\n"
                "\n"
                "Please check your parameters and try again."
            )
            sys.exit(1)

        return model_service_resp.name, service_type

    else:
        service_list_resp: ModelserviceModelServiceListResponse = vessl_api.model_service_list_api(
            organization_name=vessl_api.organization.name,
        )
        service_list: List[ResponseModelServiceInfo] = service_list_resp.results
        service_display_texts = [
            f"{s.name} (status: {str(s.status).capitalize()}; type: {str(s.type).capitalize()})"
            for s in service_list
        ]

        if no_prompt:
            print_error("Error: service name is required.")
            print()
            print_info("NOTE: Available services are:")
            example_count = 5
            for txt in service_display_texts[:example_count]:
                print_info(f"    - {txt}")
            if example_count < len(service_list):
                print_info(f"    - ({len(service_list)-example_count} more...)")
            sys.exit(1)

        chosen_service: Optional[ResponseModelServiceInfo] = prompt_choices(
            "Select service, or create a new one.",
            [("(Create a new service...)", None)]
            + [(txt, s) for txt, s in zip(service_display_texts, service_list)],
        )
        if chosen_service is None:
            return _create_service(
                service_name=None,
                cluster_name=cluster_name,
                serverless=serverless,
                node_port=node_port,
            )
        else:
            return chosen_service.name, ModelServiceType(chosen_service.type)


def _create_service(
    service_name: Optional[str], cluster_name: str, serverless: bool, node_port: int = 0, subdomain: str = ""
) -> Tuple[str, ModelServiceType]:
    """
    Prompts the user to help create a Service.

    Arguments:
        service_name (str | None):
            Name of the service, if known.

        cluster_name (str):
            Name of the cluster, hinted in YAML. This is used to create the service.

        serverless (bool):
            Whether to create the new service as serverless mode.
            Only present if the user explicitly provided an option for it.

        node_port (int):
            NodePort number to expose the service.
            Only present if the cluster is configured with a NodePort.

        subdomain (str):
            Subdomain to expose the service.
            Only present if the user explicitly provided an option for it.

    Returns:
        str: Name of the new Service.
        ModelServiceType: Type of the new Service.
    """

    if service_name is None:
        service_name = prompt_text("Name for new service")

    service_type: ModelServiceType
    service_type = ModelServiceType.SERVERLESS if serverless else ModelServiceType.PROVISIONED
    print_info(f"Creating service {service_name} in {service_type.value.capitalize()} mode.")

    # We only have the cluster's name; we need to resolve its ID.
    all_clusters = list_clusters()
    matching_cluster_list = [c for c in all_clusters if c.name == cluster_name]
    if len(matching_cluster_list) != 1:
        print_error(f"Error: cluster '{cluster_name}' not found!")
        print()
        print_error(
            "There might be a typo, or the cluster might not be connected to this organization."
        )
        print_error("Please check your cluster and try again.")
        print()
        print_info(
            f"NOTE: there are {len(all_clusters)} cluster(s) connected to this organization:"
        )
        for c in all_clusters:
            print_info(f"    - {c.name} (status: {str(c.status).capitalize()})")
        sys.exit(1)
    cluster = matching_cluster_list[0]

    if cluster.custom_endpoint_type == ClusterCustomEndpointType.NODE_PORT and node_port == 0:
        print_error(f"Error: cluster '{cluster_name}' is not configured with a NodePort.")
        print_error("Please configure the cluster with a NodePort and try again.")
        sys.exit(1)

    if cluster.custom_endpoint_type == ClusterCustomEndpointType.SUBDOMAIN and subdomain == "":
        print_error(f"Error: cluster '{cluster_name}' is not configured with a subdomain.")
        print_error("Please configure the cluster with a subdomain and try again.")
        sys.exit(1)

    try:
        vessl_api.model_service_create_api(
            organization_name=vessl_api.organization.name,
            model_service_create_api_input=ModelServiceCreateAPIInput(
                name=service_name, kernel_cluster_id=cluster.id, service_type=service_type.value, node_port=node_port, subdomain=subdomain
            ),
        )
    except VesslApiException as e:
        print_error("Failed to create service")
        print_error_result(e.message)
        sys.exit(1)

    print_info(f"Successfully created service: {service_name}")

    return service_name, service_type



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmRunSpecEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cluster': 'OrmKernelCluster',
        'image': 'OrmKernelImage',
        'organization': 'OrmOrganization',
        'pipeline_step_specs': 'list[OrmPipelineStepSpec]',
        'resource_spec_preset': 'OrmKernelResourceSpec',
        'run_executions': 'list[OrmRunExecution]'
    }

    attribute_map = {
        'cluster': 'cluster',
        'image': 'image',
        'organization': 'organization',
        'pipeline_step_specs': 'pipeline_step_specs',
        'resource_spec_preset': 'resource_spec_preset',
        'run_executions': 'run_executions'
    }

    def __init__(self, cluster=None, image=None, organization=None, pipeline_step_specs=None, resource_spec_preset=None, run_executions=None, local_vars_configuration=None):  # noqa: E501
        """OrmRunSpecEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._cluster = None
        self._image = None
        self._organization = None
        self._pipeline_step_specs = None
        self._resource_spec_preset = None
        self._run_executions = None
        self.discriminator = None

        if cluster is not None:
            self.cluster = cluster
        if image is not None:
            self.image = image
        if organization is not None:
            self.organization = organization
        if pipeline_step_specs is not None:
            self.pipeline_step_specs = pipeline_step_specs
        if resource_spec_preset is not None:
            self.resource_spec_preset = resource_spec_preset
        if run_executions is not None:
            self.run_executions = run_executions

    @property
    def cluster(self):
        """Gets the cluster of this OrmRunSpecEdges.  # noqa: E501


        :return: The cluster of this OrmRunSpecEdges.  # noqa: E501
        :rtype: OrmKernelCluster
        """
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        """Sets the cluster of this OrmRunSpecEdges.


        :param cluster: The cluster of this OrmRunSpecEdges.  # noqa: E501
        :type cluster: OrmKernelCluster
        """

        self._cluster = cluster

    @property
    def image(self):
        """Gets the image of this OrmRunSpecEdges.  # noqa: E501


        :return: The image of this OrmRunSpecEdges.  # noqa: E501
        :rtype: OrmKernelImage
        """
        return self._image

    @image.setter
    def image(self, image):
        """Sets the image of this OrmRunSpecEdges.


        :param image: The image of this OrmRunSpecEdges.  # noqa: E501
        :type image: OrmKernelImage
        """

        self._image = image

    @property
    def organization(self):
        """Gets the organization of this OrmRunSpecEdges.  # noqa: E501


        :return: The organization of this OrmRunSpecEdges.  # noqa: E501
        :rtype: OrmOrganization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this OrmRunSpecEdges.


        :param organization: The organization of this OrmRunSpecEdges.  # noqa: E501
        :type organization: OrmOrganization
        """

        self._organization = organization

    @property
    def pipeline_step_specs(self):
        """Gets the pipeline_step_specs of this OrmRunSpecEdges.  # noqa: E501


        :return: The pipeline_step_specs of this OrmRunSpecEdges.  # noqa: E501
        :rtype: list[OrmPipelineStepSpec]
        """
        return self._pipeline_step_specs

    @pipeline_step_specs.setter
    def pipeline_step_specs(self, pipeline_step_specs):
        """Sets the pipeline_step_specs of this OrmRunSpecEdges.


        :param pipeline_step_specs: The pipeline_step_specs of this OrmRunSpecEdges.  # noqa: E501
        :type pipeline_step_specs: list[OrmPipelineStepSpec]
        """

        self._pipeline_step_specs = pipeline_step_specs

    @property
    def resource_spec_preset(self):
        """Gets the resource_spec_preset of this OrmRunSpecEdges.  # noqa: E501


        :return: The resource_spec_preset of this OrmRunSpecEdges.  # noqa: E501
        :rtype: OrmKernelResourceSpec
        """
        return self._resource_spec_preset

    @resource_spec_preset.setter
    def resource_spec_preset(self, resource_spec_preset):
        """Sets the resource_spec_preset of this OrmRunSpecEdges.


        :param resource_spec_preset: The resource_spec_preset of this OrmRunSpecEdges.  # noqa: E501
        :type resource_spec_preset: OrmKernelResourceSpec
        """

        self._resource_spec_preset = resource_spec_preset

    @property
    def run_executions(self):
        """Gets the run_executions of this OrmRunSpecEdges.  # noqa: E501


        :return: The run_executions of this OrmRunSpecEdges.  # noqa: E501
        :rtype: list[OrmRunExecution]
        """
        return self._run_executions

    @run_executions.setter
    def run_executions(self, run_executions):
        """Sets the run_executions of this OrmRunSpecEdges.


        :param run_executions: The run_executions of this OrmRunSpecEdges.  # noqa: E501
        :type run_executions: list[OrmRunExecution]
        """

        self._run_executions = run_executions

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmRunSpecEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmRunSpecEdges):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class ResponsePipelineSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'editable': 'bool',
        'outputs': 'V1Output',
        'parameter_specs': 'dict[str, V1SingleVariableSpec]',
        'revision': 'int',
        'step_dependencies': 'list[ResponsePipelineStepDependency]',
        'steps': 'list[ResponsePipelineStep]'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'editable': 'editable',
        'outputs': 'outputs',
        'parameter_specs': 'parameter_specs',
        'revision': 'revision',
        'step_dependencies': 'step_dependencies',
        'steps': 'steps'
    }

    def __init__(self, created_dt=None, editable=None, outputs=None, parameter_specs=None, revision=None, step_dependencies=None, steps=None, local_vars_configuration=None):  # noqa: E501
        """ResponsePipelineSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._editable = None
        self._outputs = None
        self._parameter_specs = None
        self._revision = None
        self._step_dependencies = None
        self._steps = None
        self.discriminator = None

        self.created_dt = created_dt
        self.editable = editable
        if outputs is not None:
            self.outputs = outputs
        self.parameter_specs = parameter_specs
        self.revision = revision
        self.step_dependencies = step_dependencies
        self.steps = steps

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponsePipelineSpec.  # noqa: E501


        :return: The created_dt of this ResponsePipelineSpec.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponsePipelineSpec.


        :param created_dt: The created_dt of this ResponsePipelineSpec.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def editable(self):
        """Gets the editable of this ResponsePipelineSpec.  # noqa: E501


        :return: The editable of this ResponsePipelineSpec.  # noqa: E501
        :rtype: bool
        """
        return self._editable

    @editable.setter
    def editable(self, editable):
        """Sets the editable of this ResponsePipelineSpec.


        :param editable: The editable of this ResponsePipelineSpec.  # noqa: E501
        :type editable: bool
        """
        if self.local_vars_configuration.client_side_validation and editable is None:  # noqa: E501
            raise ValueError("Invalid value for `editable`, must not be `None`")  # noqa: E501

        self._editable = editable

    @property
    def outputs(self):
        """Gets the outputs of this ResponsePipelineSpec.  # noqa: E501


        :return: The outputs of this ResponsePipelineSpec.  # noqa: E501
        :rtype: V1Output
        """
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        """Sets the outputs of this ResponsePipelineSpec.


        :param outputs: The outputs of this ResponsePipelineSpec.  # noqa: E501
        :type outputs: V1Output
        """

        self._outputs = outputs

    @property
    def parameter_specs(self):
        """Gets the parameter_specs of this ResponsePipelineSpec.  # noqa: E501


        :return: The parameter_specs of this ResponsePipelineSpec.  # noqa: E501
        :rtype: dict[str, V1SingleVariableSpec]
        """
        return self._parameter_specs

    @parameter_specs.setter
    def parameter_specs(self, parameter_specs):
        """Sets the parameter_specs of this ResponsePipelineSpec.


        :param parameter_specs: The parameter_specs of this ResponsePipelineSpec.  # noqa: E501
        :type parameter_specs: dict[str, V1SingleVariableSpec]
        """
        if self.local_vars_configuration.client_side_validation and parameter_specs is None:  # noqa: E501
            raise ValueError("Invalid value for `parameter_specs`, must not be `None`")  # noqa: E501

        self._parameter_specs = parameter_specs

    @property
    def revision(self):
        """Gets the revision of this ResponsePipelineSpec.  # noqa: E501


        :return: The revision of this ResponsePipelineSpec.  # noqa: E501
        :rtype: int
        """
        return self._revision

    @revision.setter
    def revision(self, revision):
        """Sets the revision of this ResponsePipelineSpec.


        :param revision: The revision of this ResponsePipelineSpec.  # noqa: E501
        :type revision: int
        """
        if self.local_vars_configuration.client_side_validation and revision is None:  # noqa: E501
            raise ValueError("Invalid value for `revision`, must not be `None`")  # noqa: E501

        self._revision = revision

    @property
    def step_dependencies(self):
        """Gets the step_dependencies of this ResponsePipelineSpec.  # noqa: E501


        :return: The step_dependencies of this ResponsePipelineSpec.  # noqa: E501
        :rtype: list[ResponsePipelineStepDependency]
        """
        return self._step_dependencies

    @step_dependencies.setter
    def step_dependencies(self, step_dependencies):
        """Sets the step_dependencies of this ResponsePipelineSpec.


        :param step_dependencies: The step_dependencies of this ResponsePipelineSpec.  # noqa: E501
        :type step_dependencies: list[ResponsePipelineStepDependency]
        """
        if self.local_vars_configuration.client_side_validation and step_dependencies is None:  # noqa: E501
            raise ValueError("Invalid value for `step_dependencies`, must not be `None`")  # noqa: E501

        self._step_dependencies = step_dependencies

    @property
    def steps(self):
        """Gets the steps of this ResponsePipelineSpec.  # noqa: E501


        :return: The steps of this ResponsePipelineSpec.  # noqa: E501
        :rtype: list[ResponsePipelineStep]
        """
        return self._steps

    @steps.setter
    def steps(self, steps):
        """Sets the steps of this ResponsePipelineSpec.


        :param steps: The steps of this ResponsePipelineSpec.  # noqa: E501
        :type steps: list[ResponsePipelineStep]
        """
        if self.local_vars_configuration.client_side_validation and steps is None:  # noqa: E501
            raise ValueError("Invalid value for `steps`, must not be `None`")  # noqa: E501

        self._steps = steps

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponsePipelineSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponsePipelineSpec):
            return True

        return self.to_dict() != other.to_dict()



from datetime import datetime
from typing import Any, Callable, Dict, List, Union

import click
from tabulate import tabulate
from terminaltables import AsciiTable

from vessl.openapi_client.models import StorageFile
from vessl.util.fmt import format_key, format_data, format_size, format_bool

UNDEFINED = click.style("undefined", fg="red")


def print_table(objects: List[Any], keys: List[str], data_func: Callable, **kwargs) -> None:
    table_data = [[format_key(x) for x in keys]]
    table_data.extend(data_func(x) for x in objects)

    table = AsciiTable(table_data)
    table.inner_column_border = kwargs.get("inner_column_border", False)
    table.inner_heading_row_border = kwargs.get("inner_heading_row_border", False)
    table.inner_footing_row_border = kwargs.get("inner_footing_row_border", False)
    table.outer_border = kwargs.get("outer_border", False)

    print(table.table)

def print_table_tabulate(objects: List[dict]) -> None:
    print(tabulate(objects, headers="keys", tablefmt="grid"))


def print_data(data: Dict[str, Any]) -> None:
    lines = format_data(data, 0)
    print("\n".join(lines))


def print_logs(logs: List[Any]):
    timezone = datetime.now().astimezone().tzinfo
    for log in logs:
        ts = datetime.fromtimestamp(log.timestamp, tz=timezone).strftime("%H:%M:%S.%f")
        message = log.message.replace("\\r", "\n").replace("\\n", "\n")
        for x in message.split("\n"):
            print(f"[{ts}] {x}")


def truncate_datetime(value: datetime) -> Union[str, datetime]:
    if not value:
        return UNDEFINED
    return value.replace(microsecond=0)


def print_volume_files(
    files: List[StorageFile],
    keys: List[str] = None,
    data_func: Callable = None,
    **kwargs,
) -> None:
    """Print volume files

    Separately defined as a util function because it is used often across many files.
    TODO: Support recursive. Add Sort? Parse paths?
    """

    if keys is None:
        keys = ["Path", "Dir", "Size"]

    if data_func is None:
        data_func = lambda x: [
            x.path,
            format_bool(x.is_dir),
            format_size(x.size) if not x.is_dir else "-",
        ]

    print_table(files, keys, data_func)





# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmLLMDocument(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'artifact_id': 'int',
        'chunking_method_params': 'dict[str, object]',
        'chunking_method_type': 'str',
        'created_dt': 'datetime',
        'download_url': 'str',
        'edges': 'OrmLLMDocumentEdges',
        'extension': 'str',
        'filename': 'str',
        'id': 'int',
        'llm_knowledge_id': 'int',
        'organization_id': 'int',
        'parser_type': 'str',
        'size': 'str',
        'slug': 'str',
        'updated_dt': 'datetime',
        'upload_successful': 'bool',
        'uploaded_by_id': 'int'
    }

    attribute_map = {
        'artifact_id': 'artifact_id',
        'chunking_method_params': 'chunking_method_params',
        'chunking_method_type': 'chunking_method_type',
        'created_dt': 'created_dt',
        'download_url': 'download_url',
        'edges': 'edges',
        'extension': 'extension',
        'filename': 'filename',
        'id': 'id',
        'llm_knowledge_id': 'llm_knowledge_id',
        'organization_id': 'organization_id',
        'parser_type': 'parser_type',
        'size': 'size',
        'slug': 'slug',
        'updated_dt': 'updated_dt',
        'upload_successful': 'upload_successful',
        'uploaded_by_id': 'uploaded_by_id'
    }

    def __init__(self, artifact_id=None, chunking_method_params=None, chunking_method_type=None, created_dt=None, download_url=None, edges=None, extension=None, filename=None, id=None, llm_knowledge_id=None, organization_id=None, parser_type=None, size=None, slug=None, updated_dt=None, upload_successful=None, uploaded_by_id=None, local_vars_configuration=None):  # noqa: E501
        """OrmLLMDocument - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._artifact_id = None
        self._chunking_method_params = None
        self._chunking_method_type = None
        self._created_dt = None
        self._download_url = None
        self._edges = None
        self._extension = None
        self._filename = None
        self._id = None
        self._llm_knowledge_id = None
        self._organization_id = None
        self._parser_type = None
        self._size = None
        self._slug = None
        self._updated_dt = None
        self._upload_successful = None
        self._uploaded_by_id = None
        self.discriminator = None

        if artifact_id is not None:
            self.artifact_id = artifact_id
        if chunking_method_params is not None:
            self.chunking_method_params = chunking_method_params
        if chunking_method_type is not None:
            self.chunking_method_type = chunking_method_type
        if created_dt is not None:
            self.created_dt = created_dt
        if download_url is not None:
            self.download_url = download_url
        if edges is not None:
            self.edges = edges
        if extension is not None:
            self.extension = extension
        if filename is not None:
            self.filename = filename
        if id is not None:
            self.id = id
        if llm_knowledge_id is not None:
            self.llm_knowledge_id = llm_knowledge_id
        if organization_id is not None:
            self.organization_id = organization_id
        if parser_type is not None:
            self.parser_type = parser_type
        if size is not None:
            self.size = size
        if slug is not None:
            self.slug = slug
        if updated_dt is not None:
            self.updated_dt = updated_dt
        if upload_successful is not None:
            self.upload_successful = upload_successful
        if uploaded_by_id is not None:
            self.uploaded_by_id = uploaded_by_id

    @property
    def artifact_id(self):
        """Gets the artifact_id of this OrmLLMDocument.  # noqa: E501


        :return: The artifact_id of this OrmLLMDocument.  # noqa: E501
        :rtype: int
        """
        return self._artifact_id

    @artifact_id.setter
    def artifact_id(self, artifact_id):
        """Sets the artifact_id of this OrmLLMDocument.


        :param artifact_id: The artifact_id of this OrmLLMDocument.  # noqa: E501
        :type artifact_id: int
        """

        self._artifact_id = artifact_id

    @property
    def chunking_method_params(self):
        """Gets the chunking_method_params of this OrmLLMDocument.  # noqa: E501


        :return: The chunking_method_params of this OrmLLMDocument.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._chunking_method_params

    @chunking_method_params.setter
    def chunking_method_params(self, chunking_method_params):
        """Sets the chunking_method_params of this OrmLLMDocument.


        :param chunking_method_params: The chunking_method_params of this OrmLLMDocument.  # noqa: E501
        :type chunking_method_params: dict[str, object]
        """

        self._chunking_method_params = chunking_method_params

    @property
    def chunking_method_type(self):
        """Gets the chunking_method_type of this OrmLLMDocument.  # noqa: E501


        :return: The chunking_method_type of this OrmLLMDocument.  # noqa: E501
        :rtype: str
        """
        return self._chunking_method_type

    @chunking_method_type.setter
    def chunking_method_type(self, chunking_method_type):
        """Sets the chunking_method_type of this OrmLLMDocument.


        :param chunking_method_type: The chunking_method_type of this OrmLLMDocument.  # noqa: E501
        :type chunking_method_type: str
        """

        self._chunking_method_type = chunking_method_type

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmLLMDocument.  # noqa: E501


        :return: The created_dt of this OrmLLMDocument.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmLLMDocument.


        :param created_dt: The created_dt of this OrmLLMDocument.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def download_url(self):
        """Gets the download_url of this OrmLLMDocument.  # noqa: E501


        :return: The download_url of this OrmLLMDocument.  # noqa: E501
        :rtype: str
        """
        return self._download_url

    @download_url.setter
    def download_url(self, download_url):
        """Sets the download_url of this OrmLLMDocument.


        :param download_url: The download_url of this OrmLLMDocument.  # noqa: E501
        :type download_url: str
        """

        self._download_url = download_url

    @property
    def edges(self):
        """Gets the edges of this OrmLLMDocument.  # noqa: E501


        :return: The edges of this OrmLLMDocument.  # noqa: E501
        :rtype: OrmLLMDocumentEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmLLMDocument.


        :param edges: The edges of this OrmLLMDocument.  # noqa: E501
        :type edges: OrmLLMDocumentEdges
        """

        self._edges = edges

    @property
    def extension(self):
        """Gets the extension of this OrmLLMDocument.  # noqa: E501


        :return: The extension of this OrmLLMDocument.  # noqa: E501
        :rtype: str
        """
        return self._extension

    @extension.setter
    def extension(self, extension):
        """Sets the extension of this OrmLLMDocument.


        :param extension: The extension of this OrmLLMDocument.  # noqa: E501
        :type extension: str
        """

        self._extension = extension

    @property
    def filename(self):
        """Gets the filename of this OrmLLMDocument.  # noqa: E501


        :return: The filename of this OrmLLMDocument.  # noqa: E501
        :rtype: str
        """
        return self._filename

    @filename.setter
    def filename(self, filename):
        """Sets the filename of this OrmLLMDocument.


        :param filename: The filename of this OrmLLMDocument.  # noqa: E501
        :type filename: str
        """

        self._filename = filename

    @property
    def id(self):
        """Gets the id of this OrmLLMDocument.  # noqa: E501


        :return: The id of this OrmLLMDocument.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmLLMDocument.


        :param id: The id of this OrmLLMDocument.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def llm_knowledge_id(self):
        """Gets the llm_knowledge_id of this OrmLLMDocument.  # noqa: E501


        :return: The llm_knowledge_id of this OrmLLMDocument.  # noqa: E501
        :rtype: int
        """
        return self._llm_knowledge_id

    @llm_knowledge_id.setter
    def llm_knowledge_id(self, llm_knowledge_id):
        """Sets the llm_knowledge_id of this OrmLLMDocument.


        :param llm_knowledge_id: The llm_knowledge_id of this OrmLLMDocument.  # noqa: E501
        :type llm_knowledge_id: int
        """

        self._llm_knowledge_id = llm_knowledge_id

    @property
    def organization_id(self):
        """Gets the organization_id of this OrmLLMDocument.  # noqa: E501


        :return: The organization_id of this OrmLLMDocument.  # noqa: E501
        :rtype: int
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this OrmLLMDocument.


        :param organization_id: The organization_id of this OrmLLMDocument.  # noqa: E501
        :type organization_id: int
        """

        self._organization_id = organization_id

    @property
    def parser_type(self):
        """Gets the parser_type of this OrmLLMDocument.  # noqa: E501


        :return: The parser_type of this OrmLLMDocument.  # noqa: E501
        :rtype: str
        """
        return self._parser_type

    @parser_type.setter
    def parser_type(self, parser_type):
        """Sets the parser_type of this OrmLLMDocument.


        :param parser_type: The parser_type of this OrmLLMDocument.  # noqa: E501
        :type parser_type: str
        """

        self._parser_type = parser_type

    @property
    def size(self):
        """Gets the size of this OrmLLMDocument.  # noqa: E501


        :return: The size of this OrmLLMDocument.  # noqa: E501
        :rtype: str
        """
        return self._size

    @size.setter
    def size(self, size):
        """Sets the size of this OrmLLMDocument.


        :param size: The size of this OrmLLMDocument.  # noqa: E501
        :type size: str
        """

        self._size = size

    @property
    def slug(self):
        """Gets the slug of this OrmLLMDocument.  # noqa: E501


        :return: The slug of this OrmLLMDocument.  # noqa: E501
        :rtype: str
        """
        return self._slug

    @slug.setter
    def slug(self, slug):
        """Sets the slug of this OrmLLMDocument.


        :param slug: The slug of this OrmLLMDocument.  # noqa: E501
        :type slug: str
        """

        self._slug = slug

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmLLMDocument.  # noqa: E501


        :return: The updated_dt of this OrmLLMDocument.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmLLMDocument.


        :param updated_dt: The updated_dt of this OrmLLMDocument.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    @property
    def upload_successful(self):
        """Gets the upload_successful of this OrmLLMDocument.  # noqa: E501


        :return: The upload_successful of this OrmLLMDocument.  # noqa: E501
        :rtype: bool
        """
        return self._upload_successful

    @upload_successful.setter
    def upload_successful(self, upload_successful):
        """Sets the upload_successful of this OrmLLMDocument.


        :param upload_successful: The upload_successful of this OrmLLMDocument.  # noqa: E501
        :type upload_successful: bool
        """

        self._upload_successful = upload_successful

    @property
    def uploaded_by_id(self):
        """Gets the uploaded_by_id of this OrmLLMDocument.  # noqa: E501


        :return: The uploaded_by_id of this OrmLLMDocument.  # noqa: E501
        :rtype: int
        """
        return self._uploaded_by_id

    @uploaded_by_id.setter
    def uploaded_by_id(self, uploaded_by_id):
        """Sets the uploaded_by_id of this OrmLLMDocument.


        :param uploaded_by_id: The uploaded_by_id of this OrmLLMDocument.  # noqa: E501
        :type uploaded_by_id: int
        """

        self._uploaded_by_id = uploaded_by_id

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmLLMDocument):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmLLMDocument):
            return True

        return self.to_dict() != other.to_dict()



import warnings

import vessl
import vessl.internal
from vessl._version import __VERSION__
from vessl.util import logger
from vessl.util.constant import VESSL_LOG_LEVEL, VESSL_ENV
from requests.exceptions import RequestsDependencyWarning
from cryptography.utils import CryptographyDeprecationWarning

__version__ = __VERSION__

# Execution mode
EXEC_MODE = "SDK"

# Expose functions
init = vessl.internal.init
log = vessl.internal.log
progress = vessl.internal.progress
hp = vessl.internal.hyperparameters
finish = vessl.internal.finish

# Make API calls using `vessl_api`
vessl_api = vessl.internal.api
configure = vessl_api.configure
configure_access_token = vessl_api.configure_access_token
configure_organization = vessl_api.configure_default_organization
configure_project = vessl_api.configure_default_project

# Backward compatibility
update_access_token = vessl_api.configure_access_token
update_organization = vessl_api.configure_default_organization
update_project = vessl_api.configure_default_project

from vessl.dataset import (
    copy_dataset_volume_file,
    create_dataset,
    delete_dataset_volume_file,
    download_dataset_volume_file,
    list_dataset_volume_files,
    list_datasets,
    read_dataset,
    read_dataset_version,
    upload_dataset_volume_file,
)
from vessl.experiment import (
    create_experiment,
    download_experiment_output_files,
    list_experiment_logs,
    list_experiment_output_files,
    list_experiments,
    read_experiment,
    read_experiment_by_id,
    update_experiment,
)
from vessl.kernel_cluster import (
    create_cluster,
    delete_cluster,
    list_cluster_nodes,
    list_clusters,
    read_cluster,
    rename_cluster,
)
from vessl.kernel_image import list_kernel_images, read_kernel_image
from vessl.kernel_resource_spec import (
    list_kernel_resource_specs,
    read_kernel_resource_spec,
)
from vessl.model import (
    create_model,
    create_model_repository,
    delete_model,
    delete_model_repository,
    delete_model_volume_file,
    download_model_volume_file,
    list_model_repositories,
    list_model_volume_files,
    list_models,
    read_model,
    read_model_repository,
    update_model,
    update_model_repository,
    upload_model_volume_file,
)
from vessl.models import (
    VesslModel,
    ModelForServiceHFTransformer,
    ModelForServiceHFDiffuser,
)
from vessl.organization import (
    create_organization,
    list_organizations,
    read_organization,
)
from vessl.pipeline import (
    dispatch_pipeline,
    get_argument_value,
    get_context_variable,
    list_pipeline,
    list_pipeline_execution,
    read_pipeline,
    read_pipeline_execution,
    read_pipeline_step_execution,
    update_context_variables,
    update_pipeline_step_execution_variable,
)
from vessl.project import create_project, list_projects, read_project, delete_project
from vessl.run import (
    create_run,
    delete_run,
    list_run_logs,
    list_runs,
    read_run,
    terminate_run,
    update_run,
)
from vessl.service import RunnerBase, register_model, register_torch_model, launch_service_revision_from_yaml
from vessl.serving import (
    create_revision_from_yaml,
    list_revisions,
    read_gateway,
    read_revision,
    list_services,
    read_service,
)
from vessl.ssh_key import create_ssh_key, delete_ssh_key, list_ssh_keys
from vessl.sweep import (
    create_sweep,
    get_best_sweep_experiment,
    list_sweep_logs,
    list_sweeps,
    read_sweep,
    terminate_sweep,
)
from vessl.volume import (
    copy_volume_file,
    create_volume_file,
    delete_volume_file,
    list_volume_files,
    read_volume_file,
)
from vessl.workspace import (
    backup_workspace,
    connect_workspace_ssh,
    list_workspaces,
    read_workspace,
    delete_workspace,
    restore_workspace,
    update_vscode_remote_ssh,
)
import vessl.storage

# Expose others
from vessl.util.image import Image  # isort:skip
from vessl.util.audio import Audio  # isort:skip
from vessl.openapi_client.models import OrmSweepObjective as SweepObjective  # isort:skip
from vessl.openapi_client.models import OrmParameter as SweepParameter  # isort:skip
from vessl.openapi_client.models import OrmRange as SweepParameterRange  # isort:skip


# Suppress package deprecation warnings if prod env
if VESSL_ENV == "prod":
    warnings.simplefilter('ignore', RequestsDependencyWarning)
    warnings.simplefilter('ignore', CryptographyDeprecationWarning)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class V1WorkflowRevisionSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'edges': 'dict[str, list[V1SingleWorkflowEdge]]',
        'end_node_rendering_hints': 'V1RenderingHints',
        'kind': 'str',
        'nodes': 'dict[str, V1WorkflowNodeStruct]',
        'start_node_rendering_hints': 'V1RenderingHints',
        'variables': 'dict[str, V1VariableValue]'
    }

    attribute_map = {
        'edges': 'edges',
        'end_node_rendering_hints': 'end_node_rendering_hints',
        'kind': 'kind',
        'nodes': 'nodes',
        'start_node_rendering_hints': 'start_node_rendering_hints',
        'variables': 'variables'
    }

    def __init__(self, edges=None, end_node_rendering_hints=None, kind=None, nodes=None, start_node_rendering_hints=None, variables=None, local_vars_configuration=None):  # noqa: E501
        """V1WorkflowRevisionSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._edges = None
        self._end_node_rendering_hints = None
        self._kind = None
        self._nodes = None
        self._start_node_rendering_hints = None
        self._variables = None
        self.discriminator = None

        if edges is not None:
            self.edges = edges
        if end_node_rendering_hints is not None:
            self.end_node_rendering_hints = end_node_rendering_hints
        if kind is not None:
            self.kind = kind
        if nodes is not None:
            self.nodes = nodes
        if start_node_rendering_hints is not None:
            self.start_node_rendering_hints = start_node_rendering_hints
        if variables is not None:
            self.variables = variables

    @property
    def edges(self):
        """Gets the edges of this V1WorkflowRevisionSpec.  # noqa: E501


        :return: The edges of this V1WorkflowRevisionSpec.  # noqa: E501
        :rtype: dict[str, list[V1SingleWorkflowEdge]]
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this V1WorkflowRevisionSpec.


        :param edges: The edges of this V1WorkflowRevisionSpec.  # noqa: E501
        :type edges: dict[str, list[V1SingleWorkflowEdge]]
        """

        self._edges = edges

    @property
    def end_node_rendering_hints(self):
        """Gets the end_node_rendering_hints of this V1WorkflowRevisionSpec.  # noqa: E501


        :return: The end_node_rendering_hints of this V1WorkflowRevisionSpec.  # noqa: E501
        :rtype: V1RenderingHints
        """
        return self._end_node_rendering_hints

    @end_node_rendering_hints.setter
    def end_node_rendering_hints(self, end_node_rendering_hints):
        """Sets the end_node_rendering_hints of this V1WorkflowRevisionSpec.


        :param end_node_rendering_hints: The end_node_rendering_hints of this V1WorkflowRevisionSpec.  # noqa: E501
        :type end_node_rendering_hints: V1RenderingHints
        """

        self._end_node_rendering_hints = end_node_rendering_hints

    @property
    def kind(self):
        """Gets the kind of this V1WorkflowRevisionSpec.  # noqa: E501


        :return: The kind of this V1WorkflowRevisionSpec.  # noqa: E501
        :rtype: str
        """
        return self._kind

    @kind.setter
    def kind(self, kind):
        """Sets the kind of this V1WorkflowRevisionSpec.


        :param kind: The kind of this V1WorkflowRevisionSpec.  # noqa: E501
        :type kind: str
        """
        allowed_values = ["workflowv1/Workflow"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and kind not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `kind` ({0}), must be one of {1}"  # noqa: E501
                .format(kind, allowed_values)
            )

        self._kind = kind

    @property
    def nodes(self):
        """Gets the nodes of this V1WorkflowRevisionSpec.  # noqa: E501


        :return: The nodes of this V1WorkflowRevisionSpec.  # noqa: E501
        :rtype: dict[str, V1WorkflowNodeStruct]
        """
        return self._nodes

    @nodes.setter
    def nodes(self, nodes):
        """Sets the nodes of this V1WorkflowRevisionSpec.


        :param nodes: The nodes of this V1WorkflowRevisionSpec.  # noqa: E501
        :type nodes: dict[str, V1WorkflowNodeStruct]
        """

        self._nodes = nodes

    @property
    def start_node_rendering_hints(self):
        """Gets the start_node_rendering_hints of this V1WorkflowRevisionSpec.  # noqa: E501


        :return: The start_node_rendering_hints of this V1WorkflowRevisionSpec.  # noqa: E501
        :rtype: V1RenderingHints
        """
        return self._start_node_rendering_hints

    @start_node_rendering_hints.setter
    def start_node_rendering_hints(self, start_node_rendering_hints):
        """Sets the start_node_rendering_hints of this V1WorkflowRevisionSpec.


        :param start_node_rendering_hints: The start_node_rendering_hints of this V1WorkflowRevisionSpec.  # noqa: E501
        :type start_node_rendering_hints: V1RenderingHints
        """

        self._start_node_rendering_hints = start_node_rendering_hints

    @property
    def variables(self):
        """Gets the variables of this V1WorkflowRevisionSpec.  # noqa: E501


        :return: The variables of this V1WorkflowRevisionSpec.  # noqa: E501
        :rtype: dict[str, V1VariableValue]
        """
        return self._variables

    @variables.setter
    def variables(self, variables):
        """Sets the variables of this V1WorkflowRevisionSpec.


        :param variables: The variables of this V1WorkflowRevisionSpec.  # noqa: E501
        :type variables: dict[str, V1VariableValue]
        """

        self._variables = variables

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1WorkflowRevisionSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1WorkflowRevisionSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmRange(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'list': 'list[str]',
        'max': 'str',
        'min': 'str',
        'step': 'str'
    }

    attribute_map = {
        'list': 'list',
        'max': 'max',
        'min': 'min',
        'step': 'step'
    }

    def __init__(self, list=None, max=None, min=None, step=None, local_vars_configuration=None):  # noqa: E501
        """OrmRange - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._list = None
        self._max = None
        self._min = None
        self._step = None
        self.discriminator = None

        if list is not None:
            self.list = list
        if max is not None:
            self.max = max
        if min is not None:
            self.min = min
        if step is not None:
            self.step = step

    @property
    def list(self):
        """Gets the list of this OrmRange.  # noqa: E501


        :return: The list of this OrmRange.  # noqa: E501
        :rtype: list[str]
        """
        return self._list

    @list.setter
    def list(self, list):
        """Sets the list of this OrmRange.


        :param list: The list of this OrmRange.  # noqa: E501
        :type list: list[str]
        """

        self._list = list

    @property
    def max(self):
        """Gets the max of this OrmRange.  # noqa: E501


        :return: The max of this OrmRange.  # noqa: E501
        :rtype: str
        """
        return self._max

    @max.setter
    def max(self, max):
        """Sets the max of this OrmRange.


        :param max: The max of this OrmRange.  # noqa: E501
        :type max: str
        """

        self._max = max

    @property
    def min(self):
        """Gets the min of this OrmRange.  # noqa: E501


        :return: The min of this OrmRange.  # noqa: E501
        :rtype: str
        """
        return self._min

    @min.setter
    def min(self, min):
        """Sets the min of this OrmRange.


        :param min: The min of this OrmRange.  # noqa: E501
        :type min: str
        """

        self._min = min

    @property
    def step(self):
        """Gets the step of this OrmRange.  # noqa: E501


        :return: The step of this OrmRange.  # noqa: E501
        :rtype: str
        """
        return self._step

    @step.setter
    def step(self, step):
        """Sets the step of this OrmRange.


        :param step: The step of this OrmRange.  # noqa: E501
        :type step: str
        """

        self._step = step

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmRange):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmRange):
            return True

        return self.to_dict() != other.to_dict()



import json
import os
from typing import Callable, List, Optional

import google.auth
from google.cloud import storage as gcslib
from google.cloud.exceptions import NotFound

from vessl.util.path import ensure_parent_dir_exists

from .abstract_volume import AbstractVolume
from .file import VolumeFile


class GCSVolume(AbstractVolume):
    def __init__(
        self,
        bucket_name: str,
        base_path: str,
        gcp_api_key_json: Optional[str] = "",
    ):
        self.gcp_api_key_json = gcp_api_key_json
        self.bucket_name = bucket_name
        self.base_path = base_path
        self._init_gcs()

    def _init_gcs(self):
        gcp_api_key_dict = json.loads(self.gcp_api_key_json)
        gcp_credentials, _ = google.auth.load_credentials_from_dict(gcp_api_key_dict)
        print("gcp_credentials", gcp_credentials)
        gcs_client = gcslib.Client(credentials=gcp_credentials)

        self.gcs_client = gcs_client
        self.bucket = gcs_client.bucket(bucket_name=self.bucket_name)

    def list(self, path: str) -> List[VolumeFile]:
        path_in_bucket = os.path.join(self.base_path, path)

        blobs = self.bucket.list_blobs(
            prefix=path_in_bucket,
        )

        files = []
        # Although a one-liner is also possible, we want to benefit from type check;
        # typing of list_blobs() is not very good.
        for blob in blobs:
            blob: gcslib.Blob = blob

            # Relative path to the volume root
            relative_path = os.path.relpath(blob.name, path_in_bucket)
            files.append(VolumeFile(path=relative_path, size=blob.size))

        return files

    def download_file(
        self,
        file: VolumeFile,
        destination: str,
        progress_callback: Optional[Callable[[int], None]] = None,
    ):
        path_in_bucket = os.path.join(self.base_path, file.path)

        ensure_parent_dir_exists(destination)

        self.gcs_client.download_blob_to_file(
            self.bucket.blob(blob_name=path_in_bucket),
            file_obj=open(destination, "wb"),
        )

    def download_directory(
        self,
        directory: str,
        destination: str,
        progress_callback: Optional[Callable[[int, int, int, int, int], None]] = None,
    ):
        self._download_directory_serially(directory, destination, progress_callback)

    def upload_file(
        self,
        local_file: str,
        destination: str,
        progress_callback: Optional[Callable[[int], None]] = None,
    ):
        dest_in_bucket = os.path.join(self.base_path, destination)
        self.bucket.blob(blob_name=dest_in_bucket).upload_from_filename(
            local_file,
        )

    def delete_file(self, path: str):
        blob = os.path.join(self.base_path, path)
        # Ignore NotFound errors to align with the behavior of other volumes (e.g., S3)
        try:
            self.bucket.delete_blob(blob_name=blob)
        except NotFound:
            pass

    def delete_directory(self, path: str):
        blob = os.path.join(self.base_path, path)
        files = [os.path.join(blob, v.path) for v in self.list(path)]
        # Ignore NotFound errors to align with the behavior of other volumes (e.g., S3)
        try:
            self.bucket.delete_blobs(blobs=files)
        except NotFound:
            pass



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class ResponseVolumeV2Info(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_by_username': 'str',
        'created_dt': 'datetime',
        'id': 'int',
        'is_auto_created': 'bool',
        'name': 'str',
        'read_only': 'bool',
        'sub_path': 'str',
        'tags': 'list[ResponseTagResponse]',
        'updated_dt': 'datetime',
        'workload_info': 'ResponseWorkloadInfo'
    }

    attribute_map = {
        'created_by_username': 'created_by_username',
        'created_dt': 'created_dt',
        'id': 'id',
        'is_auto_created': 'is_auto_created',
        'name': 'name',
        'read_only': 'read_only',
        'sub_path': 'sub_path',
        'tags': 'tags',
        'updated_dt': 'updated_dt',
        'workload_info': 'workload_info'
    }

    def __init__(self, created_by_username=None, created_dt=None, id=None, is_auto_created=None, name=None, read_only=None, sub_path=None, tags=None, updated_dt=None, workload_info=None, local_vars_configuration=None):  # noqa: E501
        """ResponseVolumeV2Info - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_by_username = None
        self._created_dt = None
        self._id = None
        self._is_auto_created = None
        self._name = None
        self._read_only = None
        self._sub_path = None
        self._tags = None
        self._updated_dt = None
        self._workload_info = None
        self.discriminator = None

        self.created_by_username = created_by_username
        self.created_dt = created_dt
        self.id = id
        self.is_auto_created = is_auto_created
        self.name = name
        self.read_only = read_only
        self.sub_path = sub_path
        self.tags = tags
        self.updated_dt = updated_dt
        if workload_info is not None:
            self.workload_info = workload_info

    @property
    def created_by_username(self):
        """Gets the created_by_username of this ResponseVolumeV2Info.  # noqa: E501


        :return: The created_by_username of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: str
        """
        return self._created_by_username

    @created_by_username.setter
    def created_by_username(self, created_by_username):
        """Sets the created_by_username of this ResponseVolumeV2Info.


        :param created_by_username: The created_by_username of this ResponseVolumeV2Info.  # noqa: E501
        :type created_by_username: str
        """

        self._created_by_username = created_by_username

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseVolumeV2Info.  # noqa: E501


        :return: The created_dt of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseVolumeV2Info.


        :param created_dt: The created_dt of this ResponseVolumeV2Info.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def id(self):
        """Gets the id of this ResponseVolumeV2Info.  # noqa: E501


        :return: The id of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseVolumeV2Info.


        :param id: The id of this ResponseVolumeV2Info.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def is_auto_created(self):
        """Gets the is_auto_created of this ResponseVolumeV2Info.  # noqa: E501


        :return: The is_auto_created of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: bool
        """
        return self._is_auto_created

    @is_auto_created.setter
    def is_auto_created(self, is_auto_created):
        """Sets the is_auto_created of this ResponseVolumeV2Info.


        :param is_auto_created: The is_auto_created of this ResponseVolumeV2Info.  # noqa: E501
        :type is_auto_created: bool
        """
        if self.local_vars_configuration.client_side_validation and is_auto_created is None:  # noqa: E501
            raise ValueError("Invalid value for `is_auto_created`, must not be `None`")  # noqa: E501

        self._is_auto_created = is_auto_created

    @property
    def name(self):
        """Gets the name of this ResponseVolumeV2Info.  # noqa: E501


        :return: The name of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseVolumeV2Info.


        :param name: The name of this ResponseVolumeV2Info.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def read_only(self):
        """Gets the read_only of this ResponseVolumeV2Info.  # noqa: E501


        :return: The read_only of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: bool
        """
        return self._read_only

    @read_only.setter
    def read_only(self, read_only):
        """Sets the read_only of this ResponseVolumeV2Info.


        :param read_only: The read_only of this ResponseVolumeV2Info.  # noqa: E501
        :type read_only: bool
        """
        if self.local_vars_configuration.client_side_validation and read_only is None:  # noqa: E501
            raise ValueError("Invalid value for `read_only`, must not be `None`")  # noqa: E501

        self._read_only = read_only

    @property
    def sub_path(self):
        """Gets the sub_path of this ResponseVolumeV2Info.  # noqa: E501


        :return: The sub_path of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: str
        """
        return self._sub_path

    @sub_path.setter
    def sub_path(self, sub_path):
        """Sets the sub_path of this ResponseVolumeV2Info.


        :param sub_path: The sub_path of this ResponseVolumeV2Info.  # noqa: E501
        :type sub_path: str
        """
        if self.local_vars_configuration.client_side_validation and sub_path is None:  # noqa: E501
            raise ValueError("Invalid value for `sub_path`, must not be `None`")  # noqa: E501

        self._sub_path = sub_path

    @property
    def tags(self):
        """Gets the tags of this ResponseVolumeV2Info.  # noqa: E501


        :return: The tags of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: list[ResponseTagResponse]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """Sets the tags of this ResponseVolumeV2Info.


        :param tags: The tags of this ResponseVolumeV2Info.  # noqa: E501
        :type tags: list[ResponseTagResponse]
        """
        if self.local_vars_configuration.client_side_validation and tags is None:  # noqa: E501
            raise ValueError("Invalid value for `tags`, must not be `None`")  # noqa: E501

        self._tags = tags

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseVolumeV2Info.  # noqa: E501


        :return: The updated_dt of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseVolumeV2Info.


        :param updated_dt: The updated_dt of this ResponseVolumeV2Info.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    @property
    def workload_info(self):
        """Gets the workload_info of this ResponseVolumeV2Info.  # noqa: E501


        :return: The workload_info of this ResponseVolumeV2Info.  # noqa: E501
        :rtype: ResponseWorkloadInfo
        """
        return self._workload_info

    @workload_info.setter
    def workload_info(self, workload_info):
        """Sets the workload_info of this ResponseVolumeV2Info.


        :param workload_info: The workload_info of this ResponseVolumeV2Info.  # noqa: E501
        :type workload_info: ResponseWorkloadInfo
        """

        self._workload_info = workload_info

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseVolumeV2Info):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseVolumeV2Info):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class ResponseProjectInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'description': 'str',
        'experiment_summary': 'ResponseProjectExperimentSummary',
        'id': 'int',
        'is_example': 'bool',
        'key_metrics': 'OrmKeyMetrics',
        'name': 'str',
        'note': 'str',
        'organization': 'ResponseOrganization',
        'primary_owner': 'ResponseUser',
        'project_datasets': 'list[ResponseProjectDataset]',
        'project_repositories': 'list[ResponseProjectRepository]',
        'run_execution_summary': 'ResponseProjectRunExecutionSummary',
        'starred': 'bool',
        'sweep_summary': 'ResponseProjectSweepSummary',
        'updated_dt': 'datetime',
        'volume_id': 'int'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'description': 'description',
        'experiment_summary': 'experiment_summary',
        'id': 'id',
        'is_example': 'is_example',
        'key_metrics': 'key_metrics',
        'name': 'name',
        'note': 'note',
        'organization': 'organization',
        'primary_owner': 'primary_owner',
        'project_datasets': 'project_datasets',
        'project_repositories': 'project_repositories',
        'run_execution_summary': 'run_execution_summary',
        'starred': 'starred',
        'sweep_summary': 'sweep_summary',
        'updated_dt': 'updated_dt',
        'volume_id': 'volume_id'
    }

    def __init__(self, created_dt=None, description=None, experiment_summary=None, id=None, is_example=None, key_metrics=None, name=None, note=None, organization=None, primary_owner=None, project_datasets=None, project_repositories=None, run_execution_summary=None, starred=None, sweep_summary=None, updated_dt=None, volume_id=None, local_vars_configuration=None):  # noqa: E501
        """ResponseProjectInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._description = None
        self._experiment_summary = None
        self._id = None
        self._is_example = None
        self._key_metrics = None
        self._name = None
        self._note = None
        self._organization = None
        self._primary_owner = None
        self._project_datasets = None
        self._project_repositories = None
        self._run_execution_summary = None
        self._starred = None
        self._sweep_summary = None
        self._updated_dt = None
        self._volume_id = None
        self.discriminator = None

        self.created_dt = created_dt
        self.description = description
        self.experiment_summary = experiment_summary
        self.id = id
        self.is_example = is_example
        if key_metrics is not None:
            self.key_metrics = key_metrics
        self.name = name
        self.note = note
        self.organization = organization
        self.primary_owner = primary_owner
        if project_datasets is not None:
            self.project_datasets = project_datasets
        if project_repositories is not None:
            self.project_repositories = project_repositories
        if run_execution_summary is not None:
            self.run_execution_summary = run_execution_summary
        self.starred = starred
        if sweep_summary is not None:
            self.sweep_summary = sweep_summary
        self.updated_dt = updated_dt
        self.volume_id = volume_id

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseProjectInfo.  # noqa: E501


        :return: The created_dt of this ResponseProjectInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseProjectInfo.


        :param created_dt: The created_dt of this ResponseProjectInfo.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def description(self):
        """Gets the description of this ResponseProjectInfo.  # noqa: E501


        :return: The description of this ResponseProjectInfo.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponseProjectInfo.


        :param description: The description of this ResponseProjectInfo.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def experiment_summary(self):
        """Gets the experiment_summary of this ResponseProjectInfo.  # noqa: E501


        :return: The experiment_summary of this ResponseProjectInfo.  # noqa: E501
        :rtype: ResponseProjectExperimentSummary
        """
        return self._experiment_summary

    @experiment_summary.setter
    def experiment_summary(self, experiment_summary):
        """Sets the experiment_summary of this ResponseProjectInfo.


        :param experiment_summary: The experiment_summary of this ResponseProjectInfo.  # noqa: E501
        :type experiment_summary: ResponseProjectExperimentSummary
        """
        if self.local_vars_configuration.client_side_validation and experiment_summary is None:  # noqa: E501
            raise ValueError("Invalid value for `experiment_summary`, must not be `None`")  # noqa: E501

        self._experiment_summary = experiment_summary

    @property
    def id(self):
        """Gets the id of this ResponseProjectInfo.  # noqa: E501


        :return: The id of this ResponseProjectInfo.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseProjectInfo.


        :param id: The id of this ResponseProjectInfo.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def is_example(self):
        """Gets the is_example of this ResponseProjectInfo.  # noqa: E501


        :return: The is_example of this ResponseProjectInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_example

    @is_example.setter
    def is_example(self, is_example):
        """Sets the is_example of this ResponseProjectInfo.


        :param is_example: The is_example of this ResponseProjectInfo.  # noqa: E501
        :type is_example: bool
        """
        if self.local_vars_configuration.client_side_validation and is_example is None:  # noqa: E501
            raise ValueError("Invalid value for `is_example`, must not be `None`")  # noqa: E501

        self._is_example = is_example

    @property
    def key_metrics(self):
        """Gets the key_metrics of this ResponseProjectInfo.  # noqa: E501


        :return: The key_metrics of this ResponseProjectInfo.  # noqa: E501
        :rtype: OrmKeyMetrics
        """
        return self._key_metrics

    @key_metrics.setter
    def key_metrics(self, key_metrics):
        """Sets the key_metrics of this ResponseProjectInfo.


        :param key_metrics: The key_metrics of this ResponseProjectInfo.  # noqa: E501
        :type key_metrics: OrmKeyMetrics
        """

        self._key_metrics = key_metrics

    @property
    def name(self):
        """Gets the name of this ResponseProjectInfo.  # noqa: E501


        :return: The name of this ResponseProjectInfo.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseProjectInfo.


        :param name: The name of this ResponseProjectInfo.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def note(self):
        """Gets the note of this ResponseProjectInfo.  # noqa: E501


        :return: The note of this ResponseProjectInfo.  # noqa: E501
        :rtype: str
        """
        return self._note

    @note.setter
    def note(self, note):
        """Sets the note of this ResponseProjectInfo.


        :param note: The note of this ResponseProjectInfo.  # noqa: E501
        :type note: str
        """

        self._note = note

    @property
    def organization(self):
        """Gets the organization of this ResponseProjectInfo.  # noqa: E501


        :return: The organization of this ResponseProjectInfo.  # noqa: E501
        :rtype: ResponseOrganization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this ResponseProjectInfo.


        :param organization: The organization of this ResponseProjectInfo.  # noqa: E501
        :type organization: ResponseOrganization
        """
        if self.local_vars_configuration.client_side_validation and organization is None:  # noqa: E501
            raise ValueError("Invalid value for `organization`, must not be `None`")  # noqa: E501

        self._organization = organization

    @property
    def primary_owner(self):
        """Gets the primary_owner of this ResponseProjectInfo.  # noqa: E501


        :return: The primary_owner of this ResponseProjectInfo.  # noqa: E501
        :rtype: ResponseUser
        """
        return self._primary_owner

    @primary_owner.setter
    def primary_owner(self, primary_owner):
        """Sets the primary_owner of this ResponseProjectInfo.


        :param primary_owner: The primary_owner of this ResponseProjectInfo.  # noqa: E501
        :type primary_owner: ResponseUser
        """
        if self.local_vars_configuration.client_side_validation and primary_owner is None:  # noqa: E501
            raise ValueError("Invalid value for `primary_owner`, must not be `None`")  # noqa: E501

        self._primary_owner = primary_owner

    @property
    def project_datasets(self):
        """Gets the project_datasets of this ResponseProjectInfo.  # noqa: E501


        :return: The project_datasets of this ResponseProjectInfo.  # noqa: E501
        :rtype: list[ResponseProjectDataset]
        """
        return self._project_datasets

    @project_datasets.setter
    def project_datasets(self, project_datasets):
        """Sets the project_datasets of this ResponseProjectInfo.


        :param project_datasets: The project_datasets of this ResponseProjectInfo.  # noqa: E501
        :type project_datasets: list[ResponseProjectDataset]
        """

        self._project_datasets = project_datasets

    @property
    def project_repositories(self):
        """Gets the project_repositories of this ResponseProjectInfo.  # noqa: E501


        :return: The project_repositories of this ResponseProjectInfo.  # noqa: E501
        :rtype: list[ResponseProjectRepository]
        """
        return self._project_repositories

    @project_repositories.setter
    def project_repositories(self, project_repositories):
        """Sets the project_repositories of this ResponseProjectInfo.


        :param project_repositories: The project_repositories of this ResponseProjectInfo.  # noqa: E501
        :type project_repositories: list[ResponseProjectRepository]
        """

        self._project_repositories = project_repositories

    @property
    def run_execution_summary(self):
        """Gets the run_execution_summary of this ResponseProjectInfo.  # noqa: E501


        :return: The run_execution_summary of this ResponseProjectInfo.  # noqa: E501
        :rtype: ResponseProjectRunExecutionSummary
        """
        return self._run_execution_summary

    @run_execution_summary.setter
    def run_execution_summary(self, run_execution_summary):
        """Sets the run_execution_summary of this ResponseProjectInfo.


        :param run_execution_summary: The run_execution_summary of this ResponseProjectInfo.  # noqa: E501
        :type run_execution_summary: ResponseProjectRunExecutionSummary
        """

        self._run_execution_summary = run_execution_summary

    @property
    def starred(self):
        """Gets the starred of this ResponseProjectInfo.  # noqa: E501


        :return: The starred of this ResponseProjectInfo.  # noqa: E501
        :rtype: bool
        """
        return self._starred

    @starred.setter
    def starred(self, starred):
        """Sets the starred of this ResponseProjectInfo.


        :param starred: The starred of this ResponseProjectInfo.  # noqa: E501
        :type starred: bool
        """
        if self.local_vars_configuration.client_side_validation and starred is None:  # noqa: E501
            raise ValueError("Invalid value for `starred`, must not be `None`")  # noqa: E501

        self._starred = starred

    @property
    def sweep_summary(self):
        """Gets the sweep_summary of this ResponseProjectInfo.  # noqa: E501


        :return: The sweep_summary of this ResponseProjectInfo.  # noqa: E501
        :rtype: ResponseProjectSweepSummary
        """
        return self._sweep_summary

    @sweep_summary.setter
    def sweep_summary(self, sweep_summary):
        """Sets the sweep_summary of this ResponseProjectInfo.


        :param sweep_summary: The sweep_summary of this ResponseProjectInfo.  # noqa: E501
        :type sweep_summary: ResponseProjectSweepSummary
        """

        self._sweep_summary = sweep_summary

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseProjectInfo.  # noqa: E501


        :return: The updated_dt of this ResponseProjectInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseProjectInfo.


        :param updated_dt: The updated_dt of this ResponseProjectInfo.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    @property
    def volume_id(self):
        """Gets the volume_id of this ResponseProjectInfo.  # noqa: E501


        :return: The volume_id of this ResponseProjectInfo.  # noqa: E501
        :rtype: int
        """
        return self._volume_id

    @volume_id.setter
    def volume_id(self, volume_id):
        """Sets the volume_id of this ResponseProjectInfo.


        :param volume_id: The volume_id of this ResponseProjectInfo.  # noqa: E501
        :type volume_id: int
        """
        if self.local_vars_configuration.client_side_validation and volume_id is None:  # noqa: E501
            raise ValueError("Invalid value for `volume_id`, must not be `None`")  # noqa: E501

        self._volume_id = volume_id

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseProjectInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseProjectInfo):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import re  # noqa: F401

# python 2 and python 3 compatibility library
import six

from vessl.openapi_client.api_client import ApiClient
from vessl.openapi_client.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError
)


class SSOAPIV1Api(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def s_sosaml_initiate_login_api(self, organization_name, **kwargs):  # noqa: E501
        """s_sosaml_initiate_login_api  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.s_sosaml_initiate_login_api(organization_name, async_req=True)
        >>> result = thread.get()

        :param organization_name: (required)
        :type organization_name: str
        :param ssosaml_initiate_login_api_input:
        :type ssosaml_initiate_login_api_input: SSOSAMLInitiateLoginAPIInput
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ResponseSSOSAMLInitiateLoginResponse
        """
        kwargs['_return_http_data_only'] = True
        return self.s_sosaml_initiate_login_api_with_http_info(organization_name, **kwargs)  # noqa: E501

    def s_sosaml_initiate_login_api_with_http_info(self, organization_name, **kwargs):  # noqa: E501
        """s_sosaml_initiate_login_api  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.s_sosaml_initiate_login_api_with_http_info(organization_name, async_req=True)
        >>> result = thread.get()

        :param organization_name: (required)
        :type organization_name: str
        :param ssosaml_initiate_login_api_input:
        :type ssosaml_initiate_login_api_input: SSOSAMLInitiateLoginAPIInput
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ResponseSSOSAMLInitiateLoginResponse, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'organization_name',
            'ssosaml_initiate_login_api_input'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method s_sosaml_initiate_login_api" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'organization_name' is set
        if self.api_client.client_side_validation and local_var_params.get('organization_name') is None:  # noqa: E501
            raise ApiValueError("Missing the required parameter `organization_name` when calling `s_sosaml_initiate_login_api`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'organization_name' in local_var_params:
            path_params['organizationName'] = local_var_params['organization_name']  # noqa: E501

        query_params = []

        header_params = dict(local_var_params.get('_headers', {}))

        form_params = []
        local_var_files = {}

        body_params = None
        if 'ssosaml_initiate_login_api_input' in local_var_params:
            body_params = local_var_params['ssosaml_initiate_login_api_input']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        content_types_list = local_var_params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json'],
                'POST', body_params))  # noqa: E501
        if content_types_list:
                header_params['Content-Type'] = content_types_list

        # Authentication setting
        auth_settings = []  # noqa: E501

        response_types_map = {
            200: "ResponseSSOSAMLInitiateLoginResponse",
        }

        return self.api_client.call_api(
            '/sso/v1/{organizationName}/saml/login', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def s_sosaml_sign_in_api(self, organization_name, **kwargs):  # noqa: E501
        """s_sosaml_sign_in_api  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.s_sosaml_sign_in_api(organization_name, async_req=True)
        >>> result = thread.get()

        :param organization_name: (required)
        :type organization_name: str
        :param ssosaml_sign_in_api_input:
        :type ssosaml_sign_in_api_input: SSOSAMLSignInAPIInput
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ResponseSSOSAMLSignInResponse
        """
        kwargs['_return_http_data_only'] = True
        return self.s_sosaml_sign_in_api_with_http_info(organization_name, **kwargs)  # noqa: E501

    def s_sosaml_sign_in_api_with_http_info(self, organization_name, **kwargs):  # noqa: E501
        """s_sosaml_sign_in_api  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.s_sosaml_sign_in_api_with_http_info(organization_name, async_req=True)
        >>> result = thread.get()

        :param organization_name: (required)
        :type organization_name: str
        :param ssosaml_sign_in_api_input:
        :type ssosaml_sign_in_api_input: SSOSAMLSignInAPIInput
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ResponseSSOSAMLSignInResponse, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'organization_name',
            'ssosaml_sign_in_api_input'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method s_sosaml_sign_in_api" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'organization_name' is set
        if self.api_client.client_side_validation and local_var_params.get('organization_name') is None:  # noqa: E501
            raise ApiValueError("Missing the required parameter `organization_name` when calling `s_sosaml_sign_in_api`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'organization_name' in local_var_params:
            path_params['organizationName'] = local_var_params['organization_name']  # noqa: E501

        query_params = []

        header_params = dict(local_var_params.get('_headers', {}))

        form_params = []
        local_var_files = {}

        body_params = None
        if 'ssosaml_sign_in_api_input' in local_var_params:
            body_params = local_var_params['ssosaml_sign_in_api_input']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        content_types_list = local_var_params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json'],
                'POST', body_params))  # noqa: E501
        if content_types_list:
                header_params['Content-Type'] = content_types_list

        # Authentication setting
        auth_settings = []  # noqa: E501

        response_types_map = {
            200: "ResponseSSOSAMLSignInResponse",
        }

        return self.api_client.call_api(
            '/sso/v1/{organizationName}/saml/signin', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def s_sosamlsp_metadata_read_api(self, organization_name, **kwargs):  # noqa: E501
        """s_sosamlsp_metadata_read_api  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.s_sosamlsp_metadata_read_api(organization_name, async_req=True)
        >>> result = thread.get()

        :param organization_name: (required)
        :type organization_name: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: SamlEntityDescriptor
        """
        kwargs['_return_http_data_only'] = True
        return self.s_sosamlsp_metadata_read_api_with_http_info(organization_name, **kwargs)  # noqa: E501

    def s_sosamlsp_metadata_read_api_with_http_info(self, organization_name, **kwargs):  # noqa: E501
        """s_sosamlsp_metadata_read_api  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.s_sosamlsp_metadata_read_api_with_http_info(organization_name, async_req=True)
        >>> result = thread.get()

        :param organization_name: (required)
        :type organization_name: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(SamlEntityDescriptor, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'organization_name'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method s_sosamlsp_metadata_read_api" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'organization_name' is set
        if self.api_client.client_side_validation and local_var_params.get('organization_name') is None:  # noqa: E501
            raise ApiValueError("Missing the required parameter `organization_name` when calling `s_sosamlsp_metadata_read_api`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'organization_name' in local_var_params:
            path_params['organizationName'] = local_var_params['organization_name']  # noqa: E501

        query_params = []

        header_params = dict(local_var_params.get('_headers', {}))

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        response_types_map = {
            200: "SamlEntityDescriptor",
        }

        return self.api_client.call_api(
            '/sso/v1/{organizationName}/saml/metadata', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))



import os
import subprocess
import urllib.parse
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

import click
import humanfriendly
import inquirer
import paramiko

from vessl.openapi_client import CliWorkspaceBackupCreateAPIInput
from vessl.openapi_client.models import (
    InfluxdbWorkloadLog,
    OrmVolumeMountRequests,
    OrmWorkspacePort,
    ResponseWorkspaceDetail,
    ResponseWorkspaceList,
    WorkspaceCreateAPIInput,
)
from vessl import vessl_api
from vessl.util.echo import print_warning, print_info
from vessl.kernel_cluster import list_cluster_nodes, read_cluster
from vessl.kernel_resource_spec import (
    _configure_custom_kernel_resource_spec,
    read_kernel_resource_spec,
)
from vessl.organization import _get_organization_name
from vessl.util.common import parse_time_to_ago
from vessl.util.config import VesslConfigLoader
from vessl.util.constant import (
    SSH_CONFIG_FORMAT,
    SSH_CONFIG_PATH,
    TEMP_DIR,
    WORKSPACE_BACKUP_MAX_SIZE,
    WORKSPACE_BACKUP_MAX_SIZE_FORMATTED,
)
from vessl.util.exception import InvalidWorkspaceError, VesslException
from vessl.util.random import random_string
from vessl.util.ssh import ssh_command_from_endpoint
from vessl.util.tar import Tar
from vessl.util.zipper import Zipper
from vessl.volume import (
    _configure_volume_mount_request_datasets,
    _configure_volume_mount_request_local_files,
    copy_volume_file,
)


def read_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Read workspace in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.read_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_read_api(
        workspace_id=workspace_id, organization_name=_get_organization_name(**kwargs)
    )


def list_workspaces(
    cluster_id: int = None, statuses: List[str] = None, mine: bool = True, limit: Optional[int] = None, **kwargs
) -> List[ResponseWorkspaceList]:
    """List workspaces in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        cluster_id(int): Defaults to None.
        statuses(List[str]): A list of status filter. Defaults to None.
        mine(bool): True if list only my workspaces, False otherwise. Defaults
            to True.

    Example:
        ```python
        vessl.list_workspaces(
            cluster_id=123456,
            statuses=["running"],
        )
        ```
    """
    statuses = (
        [",".join(statuses)] if statuses else None
    )  # since openapi-generator uses repeating params instead of commas
    return vessl_api.workspace_list_api(
        organization_name=_get_organization_name(**kwargs),
        cluster=cluster_id,
        mine=mine,
        statuses=statuses,
        limit=limit,
    ).results


def create_workspace(
    name: str,
    cluster_name: str,
    cluster_node_names: List[str] = None,
    kernel_resource_spec_name: str = None,
    processor_type: str = None,
    cpu_limit: float = None,
    memory_limit: str = None,
    gpu_type: str = None,
    gpu_limit: int = None,
    kernel_image_url: str = None,
    max_hours: int = 24,
    dataset_mounts: List[str] = None,
    local_files: List[str] = None,
    use_vesslignore: bool = True,
    root_volume_size: str = "100Gi",
    ports: List[Dict[str, Any]] = None,
    init_script: str = None,
    **kwargs,
) -> ResponseWorkspaceDetail:
    """Create workspace in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        name(str): Workspace name.
        cluster_name(str): Cluster name(must be specified before other options).
        cluster_node_names(List[str]): A list of candidate cluster node names.
            Defaults to None.
        kernel_resource_spec_name(str):  Resource type to run an experiment (for
            managed cluster only). Defaults to None.
        processor_type(str) cpu or gpu (for custom cluster only). Defaults to
            None.
        cpu_limit(float): Number of vCPUs (for custom cluster only). Defaults to
            None.
        memory_limit(str): Memory limit in GiB (for custom cluster only).
            Defaults to None.
        gpu_type(str): GPU type (for custom cluster only). Defaults to None.
        gpu_limit(int): Number of GPU cores (for custom cluster only). Defaults
            to None.
        kernel_image_url(str): Kernel docker image URL. Defaults to None.
        max_hours(int): Max hours limit to run. Defaults to 24.
        dataset_mounts(List[str]): A list of dataset mounts. Defaults to None.
        local_files(List[str]): A list of local file mounts. Defaults to None.
        use_vesslignore(bool): True if local files matching glob patterns
            in .vesslignore files should be ignored. Patterns apply relative to
            the directory containing that .vesslignore file.
        root_volume_size(str): Root volume size. Defaults to "100Gi".
        ports(List[Dict[str, Any]]): Port numbers to expose. Defaults to None.
        init_script(str) Custom init script. Defaults to None.

    Example:
        ```python
        vessl.create_workspace(
            name="modern-kick",
            cluster_name="vessl-oci-sanjose",
            kernel_resource_spec_name="cpu-medium",
            kernel_image_url="public.ecr.aws/vessl/kernels:py36.full-cpu.jupyter",
        )
        ```
    """
    cluster = read_cluster(cluster_name)
    cluster_node_ids = (
        [n.id for n in list_cluster_nodes(cluster.id, **kwargs) if n.name in cluster_node_names]
        if cluster_node_names is not None
        else None
    )

    kernel_resource_spec = kernel_resource_spec_id = None
    if kernel_resource_spec_name:
        kernel_resource_spec_id = read_kernel_resource_spec(
            cluster.id,
            kernel_resource_spec_name,
        ).id
    else:
        kernel_resource_spec = _configure_custom_kernel_resource_spec(
            processor_type,
            cpu_limit,
            memory_limit,
            gpu_type,
            gpu_limit,
        )

    all_ports = [
        OrmWorkspacePort(expose_type="http", port=8888, name="jupyter"),
        OrmWorkspacePort(expose_type="tcp", port=22, name="ssh"),
    ]
    if ports is not None:
        for p in ports:
            all_ports.append(
                OrmWorkspacePort(
                    expose_type=p.get("expose_type"),
                    port=p.get("port"),
                    name=p.get("name"),
                )
            )

    requests = []
    if dataset_mounts is not None:
        requests.extend(
            _configure_volume_mount_request_datasets(
                dataset_mounts, should_add_project_datasets=False, **kwargs
            )
        )
    if local_files:
        requests.extend(
            _configure_volume_mount_request_local_files(local_files, use_vesslignore, **kwargs)
        )

    volume_mount_requests = OrmVolumeMountRequests(
        root_volume_size=root_volume_size,
        requests=requests,
    )

    return vessl_api.workspace_create_api(
        organization_name=_get_organization_name(**kwargs),
        workspace_create_api_input=WorkspaceCreateAPIInput(
            cluster_id=cluster.id,
            cluster_node_ids=cluster_node_ids,
            resource_spec=kernel_resource_spec,
            resource_spec_id=kernel_resource_spec_id,
            image_url=kernel_image_url,
            name=name,
            max_hours=max_hours,
            ports=all_ports,
            volumes=volume_mount_requests,
            init_script=init_script,
        ),
    )


def delete_workspace(
    workspace_id: int,
    **kwargs
):
    vessl_api.workspace_delete_api(
        workspace_id=workspace_id,
        organization_name=_get_organization_name(**kwargs),
    )

def list_workspace_logs(
    workspace_id: int,
    tail: int = 200,
    **kwargs,
) -> List[InfluxdbWorkloadLog]:
    """List experiment logs in the default organization. If you want to override
    the default organization, then pass `organization_name` as `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.
        tail (int): The number of lines to display from the end. Display all if
            -1. Defaults to 200.

    Example:
        ```python
         vessl.list_workspace_logs(
            workspace_id=123456,
         )
        ```
    """
    if tail == -1:
        tail = None

    return vessl_api.workspace_logs_api(
        workspace_id=workspace_id,
        organization_name=_get_organization_name(**kwargs),
        limit=tail,
    ).logs


def start_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Start the workspace container in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.start_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_start_api(
        organization_name=_get_organization_name(**kwargs), workspace_id=workspace_id
    )


def stop_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Stop the workspace container in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.stop_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_stop_api(
        organization_name=_get_organization_name(**kwargs), workspace_id=workspace_id
    )


def terminate_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Terminate the workspace container in the default organization. If you
    want to override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.terminate_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_terminate_api(
        organization_name=_get_organization_name(**kwargs), workspace_id=workspace_id
    )


def backup_workspace() -> None:
    """Backup the home directory of the workspace. This command should be called
    inside a workspace.

    Example:
        ```python
        vessl.backup_workspace()
        ```
    """
    workspace_id = VesslConfigLoader().workspace
    if workspace_id is None:
        raise InvalidWorkspaceError("Can only be called within a workspace.")

    workspace = read_workspace(workspace_id)

    filename = (
        f'workspace-backup-{datetime.utcnow().strftime("%Y%m%d%H%M%S")}-{random_string()}.tar.gz'
    )
    gzip_file_full_path = os.path.join(TEMP_DIR, filename)
    home_dir = str(Path.home())

    home_dir_size = -1
    try:
        # use -k option because there is no -b option on UNIX du
        home_dir_size_in_kilobytes = int(
            subprocess.check_output(["du", "-sk", home_dir]).decode().split("\t")[0]
        )
        home_dir_size = home_dir_size_in_kilobytes * 1024
    except Exception:
        pass

    if home_dir_size > WORKSPACE_BACKUP_MAX_SIZE:
        formatted_size = humanfriendly.format_size(home_dir_size, binary=True)
        raise VesslException(
            f"Failed: The size of {home_dir} should be less than {WORKSPACE_BACKUP_MAX_SIZE_FORMATTED} to backup. "
            f"Current size is {formatted_size}.",
        )

    print("Creating a backup file...")
    size = Tar.gzip(gzip_file_full_path, home_dir, exclude_paths=[".vessl"])
    if size > WORKSPACE_BACKUP_MAX_SIZE:
        formatted_size = humanfriendly.format_size(size, binary=True)
        raise VesslException(
            f"Failed: The size of {home_dir} should be less than {WORKSPACE_BACKUP_MAX_SIZE_FORMATTED} to backup. "
            f"Current size is {formatted_size}.",
        )

    print(f"Backup file size: {humanfriendly.format_size(size, binary=True)}")
    print("Uploading the backup file...")
    copy_volume_file(
        source_volume_id=None,
        source_path=gzip_file_full_path,
        dest_volume_id=workspace.backup_volume_id,
        dest_path=filename,
    )
    os.remove(gzip_file_full_path)

    vessl_api.cli_workspace_backup_create_api(
        workspace_id=workspace_id,
        cli_workspace_backup_create_api_input=CliWorkspaceBackupCreateAPIInput(
            filename=filename,
        ),
    )


def restore_workspace() -> None:
    """Restore the home directory from the previous backup. This command should
    be called inside a workspace.

    Example:
        ```python
        vessl.restore_workspace()
        ```
    """
    workspace_id = VesslConfigLoader().workspace
    if workspace_id is None:
        raise InvalidWorkspaceError("Can only be called within a workspace.")

    workspace = read_workspace(workspace_id=workspace_id)
    if workspace.last_backup is None:
        raise click.ClickException("This workspace does not have any backup.")

    if workspace.last_backup.filename.endswith(".zip"):
        dest_path = os.path.join(TEMP_DIR, "workspace-backup.zip")
        print("Downloading the backup file...")
        copy_volume_file(
            source_volume_id=workspace.backup_volume_id,
            source_path=workspace.last_backup.filename,
            dest_volume_id=None,
            dest_path=dest_path,
        )

        zipper = Zipper(dest_path, "r")
        size = zipper.size()
        print(f"Backup file size: {size}")
        extract_path = str(Path.home())
        if size > WORKSPACE_BACKUP_MAX_SIZE:
            extract_path = os.path.join(TEMP_DIR, "workspace-backup")
        print("Extracting...")
        zipper.extractall(extract_path)
        if size > WORKSPACE_BACKUP_MAX_SIZE:
            print(
                f"Restored to {extract_path} "
                f"since the size of the backup file is larger than {WORKSPACE_BACKUP_MAX_SIZE_FORMATTED}. "
            )
        zipper.close()
        zipper.remove()
    else:
        gzip_file_full_path = os.path.join(TEMP_DIR, "workspace-backup.tar.gz")
        print("Downloading the backup file...")
        copy_volume_file(
            source_volume_id=workspace.backup_volume_id,
            source_path=workspace.last_backup.filename,
            dest_volume_id=None,
            dest_path=gzip_file_full_path,
        )

        print("Extracting...")
        Tar.extract(gzip_file_full_path, str(Path.home()))
        os.remove(gzip_file_full_path)


def connect_workspace_ssh(private_key_path: str) -> None:
    """Connect to a running workspace via SSH.

    Args:
        private_key_path(str): SSH private key path

    Example:
        ```python
        vessl.connect_workspace_ssh(
            private_key_path="~/.ssh/key_path",
        )
        ```
    """
    running_workspaces = list_workspaces(statuses=["running"], mine=True)
    if len(running_workspaces) == 0:
        raise click.ClickException("There is no running workspace.")

    if len(running_workspaces) == 1:
        workspace = running_workspaces[0]
    else:
        workspace = inquirer.prompt(
            [
                inquirer.List(
                    "question",
                    message="Select workspace",
                    choices=[
                        (f"{w.name} (created {parse_time_to_ago(w.created_dt)})", w)
                        for w in running_workspaces
                    ],
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("question")

    cmd = ssh_command_from_endpoint(workspace.endpoints.ssh.endpoint, private_key_path)
    print_info(cmd)
    ret = os.system(cmd)
    if ret != 0:
        print_warning("Try again with the correct private key path with --key-path option.")


def update_vscode_remote_ssh(private_key_path: str) -> None:
    """Update .ssh/config file for VSCode Remote-SSH plugin.

    Args:
        private_key_path(str): SSH private key path

    Example:
        ```python
        vessl.update_vscode_remote_ssh(
            private_key_path="~/.ssh/key_path",
        )
        ```
    """
    running_workspaces = list_workspaces(statuses=["running"], mine=True)
    if len(running_workspaces) == 0:
        raise click.ClickException("There is no running workspace.")

    ssh_config = paramiko.SSHConfig()
    try:
        with open(SSH_CONFIG_PATH, "r", encoding="utf-8") as fr:
            ssh_config.parse(fr)
    except FileNotFoundError:
        pass

    host_set = ssh_config.get_hostnames()
    for workspace in running_workspaces:
        host = f"{workspace.name}-{int(workspace.created_dt.timestamp())}"
        if host in host_set:
            continue

        ssh_endpoint = urllib.parse.urlparse(workspace.endpoints.ssh.endpoint)

        config_value = SSH_CONFIG_FORMAT.format(
            host=host,
            hostname=ssh_endpoint.hostname,
            port=ssh_endpoint.port,
        )
        if private_key_path:
            config_value += f"    IdentityFile {private_key_path}\n"

        with open(SSH_CONFIG_PATH, "a", encoding="utf-8") as f:
            f.write(config_value)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class ResponseDashboardDetail(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'chart_fields': 'list[ResponseDashboardChartField]',
        'chart_sections': 'list[ResponseDashboardChartSection]',
        'created_by': 'ResponseSimpleUser',
        'filters': 'list[ResponseDashboardExperimentFilterResponse]',
        'id': 'int',
        'is_main': 'bool',
        'is_private': 'bool',
        'is_starred': 'bool',
        'last_updated_by': 'ResponseSimpleUser',
        'last_updated_dt': 'datetime',
        'name': 'str',
        'sorts': 'list[ResponseDashboardExperimentSortResponse]'
    }

    attribute_map = {
        'chart_fields': 'chart_fields',
        'chart_sections': 'chart_sections',
        'created_by': 'created_by',
        'filters': 'filters',
        'id': 'id',
        'is_main': 'is_main',
        'is_private': 'is_private',
        'is_starred': 'is_starred',
        'last_updated_by': 'last_updated_by',
        'last_updated_dt': 'last_updated_dt',
        'name': 'name',
        'sorts': 'sorts'
    }

    def __init__(self, chart_fields=None, chart_sections=None, created_by=None, filters=None, id=None, is_main=None, is_private=None, is_starred=None, last_updated_by=None, last_updated_dt=None, name=None, sorts=None, local_vars_configuration=None):  # noqa: E501
        """ResponseDashboardDetail - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._chart_fields = None
        self._chart_sections = None
        self._created_by = None
        self._filters = None
        self._id = None
        self._is_main = None
        self._is_private = None
        self._is_starred = None
        self._last_updated_by = None
        self._last_updated_dt = None
        self._name = None
        self._sorts = None
        self.discriminator = None

        self.chart_fields = chart_fields
        self.chart_sections = chart_sections
        if created_by is not None:
            self.created_by = created_by
        self.filters = filters
        self.id = id
        self.is_main = is_main
        self.is_private = is_private
        self.is_starred = is_starred
        if last_updated_by is not None:
            self.last_updated_by = last_updated_by
        self.last_updated_dt = last_updated_dt
        self.name = name
        self.sorts = sorts

    @property
    def chart_fields(self):
        """Gets the chart_fields of this ResponseDashboardDetail.  # noqa: E501


        :return: The chart_fields of this ResponseDashboardDetail.  # noqa: E501
        :rtype: list[ResponseDashboardChartField]
        """
        return self._chart_fields

    @chart_fields.setter
    def chart_fields(self, chart_fields):
        """Sets the chart_fields of this ResponseDashboardDetail.


        :param chart_fields: The chart_fields of this ResponseDashboardDetail.  # noqa: E501
        :type chart_fields: list[ResponseDashboardChartField]
        """
        if self.local_vars_configuration.client_side_validation and chart_fields is None:  # noqa: E501
            raise ValueError("Invalid value for `chart_fields`, must not be `None`")  # noqa: E501

        self._chart_fields = chart_fields

    @property
    def chart_sections(self):
        """Gets the chart_sections of this ResponseDashboardDetail.  # noqa: E501


        :return: The chart_sections of this ResponseDashboardDetail.  # noqa: E501
        :rtype: list[ResponseDashboardChartSection]
        """
        return self._chart_sections

    @chart_sections.setter
    def chart_sections(self, chart_sections):
        """Sets the chart_sections of this ResponseDashboardDetail.


        :param chart_sections: The chart_sections of this ResponseDashboardDetail.  # noqa: E501
        :type chart_sections: list[ResponseDashboardChartSection]
        """
        if self.local_vars_configuration.client_side_validation and chart_sections is None:  # noqa: E501
            raise ValueError("Invalid value for `chart_sections`, must not be `None`")  # noqa: E501

        self._chart_sections = chart_sections

    @property
    def created_by(self):
        """Gets the created_by of this ResponseDashboardDetail.  # noqa: E501


        :return: The created_by of this ResponseDashboardDetail.  # noqa: E501
        :rtype: ResponseSimpleUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ResponseDashboardDetail.


        :param created_by: The created_by of this ResponseDashboardDetail.  # noqa: E501
        :type created_by: ResponseSimpleUser
        """

        self._created_by = created_by

    @property
    def filters(self):
        """Gets the filters of this ResponseDashboardDetail.  # noqa: E501


        :return: The filters of this ResponseDashboardDetail.  # noqa: E501
        :rtype: list[ResponseDashboardExperimentFilterResponse]
        """
        return self._filters

    @filters.setter
    def filters(self, filters):
        """Sets the filters of this ResponseDashboardDetail.


        :param filters: The filters of this ResponseDashboardDetail.  # noqa: E501
        :type filters: list[ResponseDashboardExperimentFilterResponse]
        """
        if self.local_vars_configuration.client_side_validation and filters is None:  # noqa: E501
            raise ValueError("Invalid value for `filters`, must not be `None`")  # noqa: E501

        self._filters = filters

    @property
    def id(self):
        """Gets the id of this ResponseDashboardDetail.  # noqa: E501


        :return: The id of this ResponseDashboardDetail.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseDashboardDetail.


        :param id: The id of this ResponseDashboardDetail.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def is_main(self):
        """Gets the is_main of this ResponseDashboardDetail.  # noqa: E501


        :return: The is_main of this ResponseDashboardDetail.  # noqa: E501
        :rtype: bool
        """
        return self._is_main

    @is_main.setter
    def is_main(self, is_main):
        """Sets the is_main of this ResponseDashboardDetail.


        :param is_main: The is_main of this ResponseDashboardDetail.  # noqa: E501
        :type is_main: bool
        """
        if self.local_vars_configuration.client_side_validation and is_main is None:  # noqa: E501
            raise ValueError("Invalid value for `is_main`, must not be `None`")  # noqa: E501

        self._is_main = is_main

    @property
    def is_private(self):
        """Gets the is_private of this ResponseDashboardDetail.  # noqa: E501


        :return: The is_private of this ResponseDashboardDetail.  # noqa: E501
        :rtype: bool
        """
        return self._is_private

    @is_private.setter
    def is_private(self, is_private):
        """Sets the is_private of this ResponseDashboardDetail.


        :param is_private: The is_private of this ResponseDashboardDetail.  # noqa: E501
        :type is_private: bool
        """
        if self.local_vars_configuration.client_side_validation and is_private is None:  # noqa: E501
            raise ValueError("Invalid value for `is_private`, must not be `None`")  # noqa: E501

        self._is_private = is_private

    @property
    def is_starred(self):
        """Gets the is_starred of this ResponseDashboardDetail.  # noqa: E501


        :return: The is_starred of this ResponseDashboardDetail.  # noqa: E501
        :rtype: bool
        """
        return self._is_starred

    @is_starred.setter
    def is_starred(self, is_starred):
        """Sets the is_starred of this ResponseDashboardDetail.


        :param is_starred: The is_starred of this ResponseDashboardDetail.  # noqa: E501
        :type is_starred: bool
        """
        if self.local_vars_configuration.client_side_validation and is_starred is None:  # noqa: E501
            raise ValueError("Invalid value for `is_starred`, must not be `None`")  # noqa: E501

        self._is_starred = is_starred

    @property
    def last_updated_by(self):
        """Gets the last_updated_by of this ResponseDashboardDetail.  # noqa: E501


        :return: The last_updated_by of this ResponseDashboardDetail.  # noqa: E501
        :rtype: ResponseSimpleUser
        """
        return self._last_updated_by

    @last_updated_by.setter
    def last_updated_by(self, last_updated_by):
        """Sets the last_updated_by of this ResponseDashboardDetail.


        :param last_updated_by: The last_updated_by of this ResponseDashboardDetail.  # noqa: E501
        :type last_updated_by: ResponseSimpleUser
        """

        self._last_updated_by = last_updated_by

    @property
    def last_updated_dt(self):
        """Gets the last_updated_dt of this ResponseDashboardDetail.  # noqa: E501


        :return: The last_updated_dt of this ResponseDashboardDetail.  # noqa: E501
        :rtype: datetime
        """
        return self._last_updated_dt

    @last_updated_dt.setter
    def last_updated_dt(self, last_updated_dt):
        """Sets the last_updated_dt of this ResponseDashboardDetail.


        :param last_updated_dt: The last_updated_dt of this ResponseDashboardDetail.  # noqa: E501
        :type last_updated_dt: datetime
        """

        self._last_updated_dt = last_updated_dt

    @property
    def name(self):
        """Gets the name of this ResponseDashboardDetail.  # noqa: E501


        :return: The name of this ResponseDashboardDetail.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseDashboardDetail.


        :param name: The name of this ResponseDashboardDetail.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def sorts(self):
        """Gets the sorts of this ResponseDashboardDetail.  # noqa: E501


        :return: The sorts of this ResponseDashboardDetail.  # noqa: E501
        :rtype: list[ResponseDashboardExperimentSortResponse]
        """
        return self._sorts

    @sorts.setter
    def sorts(self, sorts):
        """Sets the sorts of this ResponseDashboardDetail.


        :param sorts: The sorts of this ResponseDashboardDetail.  # noqa: E501
        :type sorts: list[ResponseDashboardExperimentSortResponse]
        """
        if self.local_vars_configuration.client_side_validation and sorts is None:  # noqa: E501
            raise ValueError("Invalid value for `sorts`, must not be `None`")  # noqa: E501

        self._sorts = sorts

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseDashboardDetail):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseDashboardDetail):
            return True

        return self.to_dict() != other.to_dict()



import inspect

from urllib3.exceptions import MaxRetryError

from vessl.openapi_client import (
    ExperimentUpdateHyperparametersAPIInput,
    ResponseExperimentInfo,
)
from vessl.util import logger
from vessl.util.api import VesslApi
from vessl.util.exception import VesslApiException, suppress_sdk_exception


class HyperparametersController(object):
    def __init__(self, api: VesslApi, experiment: ResponseExperimentInfo):
        object.__setattr__(self, "hyperparameters", Hyperparameters())
        object.__setattr__(self, "_api", api)
        object.__setattr__(self, "_experiment", experiment)

    def __setattr__(self, key, value):
        self.hyperparameters.__setattr__(key, value)

    def configure(self, api, experiment):
        object.__setattr__(self, "_api", api)
        object.__setattr__(self, "_experiment", experiment)

    def update_items(self, d=None):
        self.hyperparameters.update_items(d)

    def as_list_of_dict(self):
        return self.hyperparameters.as_list_of_dict()

    @suppress_sdk_exception
    def update(self, d=None):
        self.hyperparameters.update_items(d)
        logger.debug(f"Update hyperparameters: {self.hyperparameters.as_list_of_dict()}")
        try:
            self._api.experiment_update_hyperparameters_api(
                organization_name=self._experiment.organization.name,
                project_name=self._experiment.project.name,
                experiment_number=self._experiment.number,
                experiment_update_hyperparameters_api_input=ExperimentUpdateHyperparametersAPIInput(
                    hyperparameters=self.hyperparameters.as_list_of_dict(),
                ),
            )

        except (MaxRetryError, VesslApiException) as e:
            logger.exception("Failed to update hyperparameters to experiment", exc_info=e)

        except Exception as e:
            logger.exception("Unexpected error", exc_info=e)


class Hyperparameters(object):
    def __init__(self):
        object.__setattr__(self, "_items", dict())

    def __repr__(self):
        return str(dict(self._items))

    def __setattr__(self, key, value):
        key, value = self._validate(key, value)
        self._items[key] = value

    def __getattr__(self, key):
        return self._items[key]

    def __contains__(self, key):
        return key in self._items

    def keys(self):
        return [k for k in self._items.keys()]

    def as_list_of_dict(self):
        return [{"key": k, "value": v} for k, v in self._items.items()]

    def items(self):
        return [(k, v) for k, v in self._items.items()]

    def update_items(self, d=None):
        if d is None:
            return
        if isinstance(d, dict):
            d = self._validate_dict(d)
            self._items.update(d)
        elif inspect.getmodule(d).__name__ == "argparse":
            d = vars(d)
            d = self._validate_dict(d)
            self._items.update(d)

    def _validate_dict(self, d):
        validated_d = {}
        for k, v in d.items():
            if v is None:
                continue
            k, v = self._validate(k, v)
            validated_d[k] = v
        return validated_d

    def _validate(self, key, value):
        key = key.strip()
        value = str(value)
        return key, value
